# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the bspline directory in idlutils.
"""
from warnings import warn
import numpy as np
from numpy.linalg.linalg import LinAlgError
from scipy.linalg import cholesky_banded, cho_solve_banded
from . import PydlutilsUserWarning
from .math import djs_reject, flegendre
from .trace import fchebyshev
from .uniq import uniq


class Bspline(object):
    """B-spline class.

    Functions in the idlutils bspline library are implemented as methods on this
    class.

    Parameters
    ----------
    x : :class:`numpy.ndarray`
        The data.
    nord : :class:`int`, optional
        The order of the B-spline.  Default is 4, which is cubic.
    npoly : :class:`int`, optional
        Polynomial order to fit over 2nd variable, if supplied.  If not
        supplied the order is 1.
    bkpt : :class:`numpy.ndarray`, optional
        To be documented.
    bkspread : :class:`float`, optional
        To be documented.

    Attributes
    ----------
    breakpoints
        To be documented.
    nord
        To be documented.
    npoly
        To be documented.
    mask
        To be documented.
    coeff
        To be documented.
    icoeff
        To be documented.
    xmin
        To be documented.
    xmax
        To be documented.
    funcname
        To be documented.
    """

    def __init__(self, x, nord=4, npoly=1, bkpt=None, fullbkpt=None,
                 bkspread=1.0, verbose=False, **kwargs):
        """Init creates an object whose attributes are similar to the
        structure returned by the ``create_bsplineset()`` function.
        """
        #
        # Set the breakpoints.
        #
        if fullbkpt is None:
            if bkpt is None:
                startx = x.min()
                rangex = x.max() - startx
                if 'placed' in kwargs:
                    w = ((kwargs['placed'] >= startx) &
                         (kwargs['placed'] <= startx + rangex))
                    if w.sum() < 2:
                        bkpt = np.arange(2, dtype='f') * rangex + startx
                    else:
                        bkpt = kwargs['placed'][w]
                elif 'bkspace' in kwargs:
                    nbkpts = int(rangex / kwargs['bkspace']) + 1
                    if nbkpts < 2:
                        nbkpts = 2
                    tempbkspace = rangex / float(nbkpts - 1)
                    bkpt = np.arange(nbkpts, dtype='f') * tempbkspace + startx
                elif 'nbkpts' in kwargs:
                    nbkpts = kwargs['nbkpts']
                    if nbkpts < 2:
                        nbkpts = 2
                    tempbkspace = rangex / float(nbkpts - 1)
                    bkpt = np.arange(nbkpts, dtype='f') * tempbkspace + startx
                elif 'everyn' in kwargs:
                    nx = x.size
                    nbkpts = max(nx / kwargs['everyn'], 1)
                    if nbkpts == 1:
                        xspot = [0]
                    else:
                        xspot = (nx / nbkpts) * np.arange(nbkpts)
                    bkpt = np.interp(xspot, np.arange(nx), x)
                else:
                    raise ValueError('No information for bkpts.')
            imin = bkpt.argmin()
            imax = bkpt.argmax()
            if x.min() < bkpt[imin]:
                if verbose:
                    print('Lowest breakpoint does not cover lowest x value: '
                          'changing.')
                bkpt[imin] = x.min()
            if x.max() > bkpt[imax]:
                if verbose:
                    print('Highest breakpoint does not cover highest x value: '
                          'changing.')
                bkpt[imax] = x.max()
            nshortbkpt = bkpt.size
            fullbkpt = bkpt.copy()
            if nshortbkpt == 1:
                bkspace = np.float32(bkspread)
            else:
                bkspace = (bkpt[1] - bkpt[0]) * np.float32(bkspread)
            for i in np.arange(1, nord, dtype=np.float32):
                fullbkpt = np.insert(fullbkpt, 0, bkpt[0] - bkspace * i)
                fullbkpt = np.insert(fullbkpt, fullbkpt.shape[0],
                                     bkpt[nshortbkpt - 1] + bkspace * i)
        #
        # Set the attributes
        #
        nc = fullbkpt.size - nord
        self.breakpoints = fullbkpt.astype(float)
        self.nord = nord
        self.npoly = npoly
        self.mask = np.ones((fullbkpt.size,), dtype=bool)
        if npoly > 1:
            self.coeff = np.zeros((npoly, nc), dtype=float)
            self.icoeff = np.zeros((npoly, nc), dtype=float)
        else:
            self.coeff = np.zeros((nc,), dtype=float)
            self.icoeff = np.zeros((nc,), dtype=float)
        self.xmin = 0.0
        self.xmax = 1.0
        self.funcname = kwargs[
            'funcname'] if 'funcname' in kwargs else 'legendre'

        return

    def fit(self, xdata, ydata, invvar, x2=None):
        """Calculate a B-spline in the least-squares sense.

        Fit is based on two variables: `xdata` sorted and spans a large range
        where breakpoints are required `ydata` can be described with a low order
        polynomial.

        Parameters
        ----------
        xdata : :class:`numpy.ndarray`
            Independent variable.
        ydata : :class:`numpy.ndarray`
            Dependent variable.
        invvar : :class:`numpy.ndarray`
            Inverse variance of `ydata`.
        x2 : :class:`numpy.ndarray`, optional
            Orthogonal dependent variable for 2d fits.

        Returns
        -------
        :func:`tuple`
            A tuple containing an integer error code, and the evaluation of the
            b-spline at the input values.  An error code of -2 is a failure,
            -1 indicates dropped breakpoints, 0 is success, and positive
            integers indicate ill-conditioned breakpoints.
        """
        goodbk = self.mask[self.nord:]
        nn = goodbk.sum()
        if nn < self.nord:
            yfit = np.zeros(ydata.shape, dtype='f')
            return -2, yfit
        nfull = nn * self.npoly
        bw = self.npoly * self.nord
        a1, lower, upper = self.action(xdata, x2=x2)
        foo = np.tile(invvar, bw).reshape(bw, invvar.size).transpose()
        a2 = a1 * foo
        alpha = np.zeros((bw, nfull+bw), dtype='d')
        beta = np.zeros((nfull+bw,), dtype='d')
        bi = np.arange(bw, dtype='i4')
        bo = np.arange(bw, dtype='i4')
        for k in range(1, bw):
            bi = np.append(bi, np.arange(bw-k, dtype='i4')+(bw+1)*k)
            bo = np.append(bo, np.arange(bw-k, dtype='i4')+bw*k)
        for k in range(nn-self.nord+1):
            itop = k*self.npoly
            ibottom = min(itop, nfull) + bw - 1
            ict = upper[k] - lower[k] + 1
            if ict > 0:
                work = np.dot(a1[lower[k]:upper[k]+1, :].T,
                              a2[lower[k]:upper[k]+1, :])
                wb = np.dot(ydata[lower[k]:upper[k]+1],
                            a2[lower[k]:upper[k]+1, :])
                alpha.T.flat[bo+itop*bw] += work.flat[bi]
                beta[itop:ibottom+1] += wb
        min_influence = 1.0e-10 * invvar.sum() / nfull
        errb = cholesky_band(alpha, mininf=min_influence)
        if isinstance(errb[0], int) and errb[0] == -1:
            a = errb[1]
        else:
            yfit, foo = self.value(xdata, x2=x2, action=a1, upper=upper,
                                   lower=lower)
            return self.maskpoints(errb[0]), yfit
        sol = cholesky_solve(a, beta)
        if self.coeff.ndim == 2:
            # JFH made major bug fix here.
            self.icoeff[:, goodbk] = np.array(
                a[0, 0:nfull].T.reshape(self.npoly, nn, order='F'),
                dtype=a.dtype)
            self.coeff[:, goodbk] = np.array(
                sol[0:nfull].T.reshape(self.npoly, nn, order='F'),
                dtype=sol.dtype)
        else:
            self.icoeff[goodbk] = np.array(a[0, 0:nfull], dtype=a.dtype)
            self.coeff[goodbk] = np.array(sol[0:nfull], dtype=sol.dtype)
        yfit, foo = self.value(xdata, x2=x2, action=a1, upper=upper,
                               lower=lower)
        return 0, yfit

    def action(self, x, x2=None):
        """Construct banded B-spline matrix, with dimensions [ndata, bandwidth].

        Parameters
        ----------
        x : :class:`numpy.ndarray`
            Independent variable.
        x2 : :class:`numpy.ndarray`, optional
            Orthogonal dependent variable for 2d fits.

        Returns
        -------
        :func:`tuple`
            A tuple containing the B-spline action matrix; the 'lower'
            parameter, a list of pixel positions, each corresponding to the
            first occurence of position greater than breakpoint indx; and
            'upper', Same as lower, but denotes the upper pixel positions.
        """
        nx = x.size
        nbkpt = self.mask.sum()
        if nbkpt < 2*self.nord:
            return -2, 0, 0
        n = nbkpt - self.nord
        bw = self.npoly*self.nord
        lower = np.zeros((n - self.nord + 1,), dtype=int)
        upper = np.zeros((n - self.nord + 1,), dtype=int) - 1
        indx = self.intrv(x)
        bf1 = self.bsplvn(x, indx)
        action = bf1
        aa = uniq(indx, np.arange(indx.size, dtype=int))
        upper[indx[aa]-self.nord+1] = aa
        rindx = indx[::-1]
        bb = uniq(rindx, np.arange(rindx.size, dtype=int))
        lower[rindx[bb]-self.nord+1] = nx - bb - 1
        if x2 is not None:
            if x2.size != nx:
                raise ValueError('Dimensions of x and x2 do not match.')
            x2norm = 2.0 * (x2 - self.xmin) / (self.xmax - self.xmin) - 1.0
            if self.funcname == 'poly':
                temppoly = np.ones((nx, self.npoly), dtype=float)
                for i in range(1, self.npoly):
                    temppoly[:, i] = temppoly[:, i-1] * x2norm
            elif self.funcname == 'poly1':
                temppoly = np.tile(x2norm, self.npoly).reshape(nx, self.npoly)
                for i in range(1, self.npoly):
                    temppoly[:, i] = temppoly[:, i-1] * x2norm
            elif self.funcname == 'chebyshev':
                temppoly = fchebyshev(x2norm, self.npoly).T
            elif self.funcname == 'legendre':
                temppoly = flegendre(x2norm, self.npoly).T
            else:
                raise ValueError('Unknown value of funcname.')
            action = np.zeros((nx, bw), dtype=float, order='F')
            counter = -1
            for ii in range(self.nord):
                for jj in range(self.npoly):
                    counter += 1
                    action[:, counter] = bf1[:, ii]*temppoly[:, jj]
        return action, lower, upper

    def intrv(self, x):
        """Find the segment between breakpoints which contain each value
        in the array `x`.

        The minimum breakpoint is ``nbkptord - 1``, and the maximum
        is ``nbkpt - nbkptord - 1``.

        Parameters
        ----------
        x : :class:`numpy.ndarray`
            Data values, assumed to be monotonically increasing.

        Returns
        -------
        :class:`numpy.ndarray`
            Position of array elements with respect to breakpoints.
        """
        gb = self.breakpoints[self.mask]
        n = gb.size - self.nord
        indx = np.zeros((x.size,), dtype='i4')
        ileft = self.nord - 1
        for i in range(x.size):
            while x[i] > gb[ileft+1] and ileft < n - 1:
                ileft += 1
            indx[i] = ileft
        return indx

    def bsplvn(self, x, ileft):
        """Calculates the value of all possibly nonzero B-splines at `x`
        of a certain order.

        Parameters
        ----------
        x : :class:`numpy.ndarray`
            Independent variable.
        ileft : :class:`int`
            Breakpoint segements that contain `x`.

        Returns
        -------
        :class:`numpy.ndarray`
            B-spline values.
        """
        bkpt = self.breakpoints[self.mask]
        vnikx = np.zeros((x.size, self.nord), dtype=x.dtype, order='F')
        deltap = vnikx.copy()
        deltam = vnikx.copy()
        j = 0
        vnikx[:, 0] = 1.0
        while j < self.nord - 1:
            ipj = ileft+j+1
            deltap[:, j] = bkpt[ipj] - x
            imj = ileft-j
            deltam[:, j] = x - bkpt[imj]
            vmprev = 0.0
            for k in range(j+1):
                vm = vnikx[:, k]/(deltap[:, k] + deltam[:, j-k])
                vnikx[:, k] = vm*deltap[:, k] + vmprev
                vmprev = vm*deltam[:, j-k]
            j += 1
            vnikx[:, j] = vmprev
        return vnikx

    def value(self, x, x2=None, action=None, lower=None, upper=None):
        """Evaluate a B-spline at specified values.

        Parameters
        ----------
        x : :class:`numpy.ndarray`
            Independent variable.
        x2 : :class:`numpy.ndarray`, optional
            Orthogonal dependent variable for 2d fits.
        action : :class:`numpy.ndarray`, optional
            Action matrix to use.  If not supplied it is calculated.
        lower : :class:`numpy.ndarray`, optional
            If the action parameter is supplied, this parameter must also
            be supplied.
        upper : :class:`numpy.ndarray`, optional
            If the action parameter is supplied, this parameter must also
            be supplied.

        Returns
        -------
        :func:`tuple`
            A tuple containing the results of the bspline evaluation and a
            mask indicating where the evaluation was good.
        """
        xsort = x.argsort()
        xwork = x[xsort]
        if x2 is not None:
            x2work = x2[xsort]
        else:
            x2work = None
        if action is not None:
            if lower is None or upper is None:
                raise ValueError('Must specify lower and upper '
                                 'if action is set.')
        else:
            action, lower, upper = self.action(xwork, x2=x2work)
        yfit = np.zeros(x.shape, dtype=x.dtype)
        bw = self.npoly * self.nord
        spot = np.arange(bw, dtype='i4')
        goodbk = self.mask.nonzero()[0]
        coeffbk = self.mask[self.nord:].nonzero()[0]
        n = self.mask.sum() - self.nord
        if self.npoly > 1:
            goodcoeff = self.coeff[:, coeffbk]
        else:
            goodcoeff = self.coeff[coeffbk]
        # maskthis = np.zeros(xwork.shape,dtype=xwork.dtype)
        for i in range(n-self.nord+1):
            ict = upper[i] - lower[i] + 1
            if ict > 0:
                yfit[lower[i]:upper[i] + 1] = np.dot(
                    action[lower[i]:upper[i] + 1, :],
                    (goodcoeff.flatten('F'))[i * self.npoly + spot])
        yy = yfit.copy()
        yy[xsort] = yfit
        mask = np.ones(x.shape, dtype='bool')
        gb = self.breakpoints[goodbk]
        outside = ((x < gb[self.nord-1]) | (x > gb[n]))
        if outside.any():
            mask[outside] = False
        hmm = ((np.diff(goodbk) > 2).nonzero())[0]
        for jj in range(hmm.size):
            inside = ((x >= self.breakpoints[goodbk[hmm[jj]]]) &
                      (x <= self.breakpoints[goodbk[hmm[jj]+1]-1]))
            if inside.any():
                mask[inside] = False
        return yy, mask

    def maskpoints(self, err):
        """Perform simple logic of which breakpoints to mask.

        Parameters
        ----------
        err : :class:`numpy.ndarray`
            The list of indexes returned by the cholesky routines.

        Returns
        -------
        :class:`int`
            An integer indicating the results of the masking.  -1 indicates
            that the error points were successfully masked.  -2 indicates
            failure; the calculation should be aborted.

        Notes
        -----
        The mask attribute is modified, assuming it is possible to create the
        mask.
        """
        nbkpt = self.mask.sum()
        goodbkpt = np.where(self.mask)[0]
        nbkpt = len(goodbkpt)
        if nbkpt <= 2*self.nord:
            return -2
        hmm = err[uniq(err//self.npoly)]//self.npoly
        n = nbkpt - self.nord
        if np.any(hmm >= n):
            return -2
        test = np.zeros(nbkpt, dtype='bool')
        for jj in range(-int(np.ceil(self.nord/2)), int(self.nord/2.)):
            foo = np.where((hmm+jj) > 0, hmm+jj, np.zeros(hmm.shape,
                                                          dtype=hmm.dtype))
            inside = np.where((foo+self.nord) < n-1, foo+self.nord,
                              np.zeros(hmm.shape, dtype=hmm.dtype)+n-1)
            if len(inside) > 0:
                test[inside] = True
        if test.any():
            reality = goodbkpt[test]
            if self.mask[reality].any():
                self.mask[reality] = False
                return -1
            else:
                return -2
        else:
            return -2

    def workit(self, xdata, ydata, invvar, action, lower, upper):
        """An internal routine for bspline_extract and bspline_radial which
        solve a general banded correlation matrix which is represented by the
        variable "action".  This routine only solves the linear system once,
        and stores the coefficients in sset. A non-zero return value signifies
        a failed inversion
        Parameters
        ----------
        xdata : :class:`numpy.ndarray`
            Independent variable.
        ydata : :class:`numpy.ndarray`
            Dependent variable.
        invvar : :class:`numpy.ndarray`
            Inverse variance of `ydata`.
        action : :class:`numpy.ndarray`
            Banded correlation matrix
        lower  : :class:`numpy.ndarray`
            A list of pixel positions, each corresponding to the first
            occurence of position greater than breakpoint indx
        upper  : :class:`numpy.ndarray`
            Same as lower, but denotes the upper pixel positions
        Returns
        -------
        :func:`tuple`
            A tuple containing an integer error code, and the evaluation of the
            b-spline at the input values.  The error codes are as follows:
             -2:  is a failure,
             -1:  indicates dropped breakpoints
              0:  is success
              positive integers:  indicate ill-conditioned breakpoints.
        """
        goodbk = self.mask[self.nord:]
        nn = goodbk.sum()
        if nn < self.nord:
            yfit = np.zeros(ydata.shape, dtype=float)
            return -2, yfit
        nfull = nn * self.npoly
        bw = self.npoly * self.nord
        foo = np.sqrt(np.tile(invvar, bw).reshape(bw, invvar.size).transpose())
        a2 = action * foo
        # a2 = action*np.sqrt(np.outer(invvar,np.ones(bw)))

        alpha = np.zeros((bw, nfull + bw), dtype='d')
        beta = np.zeros((nfull + bw,), dtype='d')
        bi = np.arange(bw, dtype='i4')
        bo = np.arange(bw, dtype='i4')
        for k in range(1, bw):
            bi = np.append(bi, np.arange(bw - k, dtype='i4') + (bw + 1) * k)
            bo = np.append(bo, np.arange(bw - k, dtype='i4') + bw * k)
        for k in range(nn - self.nord + 1):
            itop = k * self.npoly
            ibottom = min(itop, nfull) + bw - 1
            ict = upper[k] - lower[k] + 1
            if ict > 0:
                work = np.dot(a2[lower[k]:upper[k] + 1, :].T,
                              a2[lower[k]:upper[k] + 1, :])
                wb = np.dot(ydata[lower[k]:upper[k] + 1] * np.sqrt(
                    invvar[lower[k]:upper[k] + 1]),
                            a2[lower[k]:upper[k] + 1, :])
                alpha.T.flat[bo + itop * bw] += work.flat[bi]
                beta[itop:ibottom + 1] += wb
        min_influence = 1.0e-10 * invvar.sum() / nfull
        # Right now we are not returning the covariance,
        # although it may arise that we should
        # covariance = alpha
        errb = cholesky_band(alpha, mininf=min_influence)  # ,verbose=True)
        if isinstance(errb[0], int) and errb[0] == -1:
            a = errb[1]
        else:
            yfit, foo = self.value(xdata, x2=xdata, action=action, upper=upper,
                                   lower=lower)
            return self.maskpoints(errb[0]), yfit
        errs = cholesky_solve(a, beta)
        if isinstance(errs[0], int) and errs[0] == -1:
            sol = errs[1]
        else:
            #
            # It is not possible for this to get called, because cholesky_solve
            # has only one return statement, & that statement guarantees that
            # errs[0] == -1
            #
            yfit, foo = self.value(xdata, x2=xdata, action=action, upper=upper,
                                   lower=lower)
            return self.maskpoints(errs[0]), yfit
        if self.coeff.ndim == 2:
            self.icoeff[:, goodbk] = np.array(
                a[0, 0:nfull].T.reshape(self.npoly, nn, order='F'),
                dtype=a.dtype)
            self.coeff[:, goodbk] = np.array(
                sol[0:nfull].T.reshape(self.npoly, nn, order='F'),
                dtype=sol.dtype)
        else:
            self.icoeff[goodbk] = np.array(a[0, 0:nfull], dtype=a.dtype)
            self.coeff[goodbk] = np.array(sol[0:nfull], dtype=sol.dtype)
        yfit, foo = self.value(xdata, x2=xdata, action=action, upper=upper,
                               lower=lower)
        return 0, yfit


def cholesky_band(ndl, mininf=0.0):
    """Compute *lower* Cholesky decomposition of a banded matrix.

    This function provides informative error messages to pass back to the
    :class:`~pydl.pydlutils.bspline.Bspline` machinery; the actual
    computation is delegated to :func:`scipy.linalg.cholesky_banded`.

    Parameters
    ----------
    ndl : :class:`numpy.ndarray`
        A matrix on which to perform the Cholesky decomposition.  The
        matrix must be in a special, *lower* form described in
        :func:`scipy.linalg.cholesky_banded`.  In addition, the input
        must be padded.  If the original, square matrix has size
        :math:`N \\times N`, and the width of the band is :math:`b`,
        `l` must be :math:`b \\times (N + b)`.
    mininf : :class:`float`, optional
        Entries in the `l` matrix are considered negative if they are less
        than this value (default 0.0).

    Returns
    -------
    :func:`tuple`
        If problems were detected, the first item will be the index or
        indexes where the problem was detected, and the second item will simply
        be the input matrix.  If no problems were detected, the first item
        will be -1, and the second item will be the Cholesky decomposition.
    """
    bw, nn = ndl.shape
    n = nn - bw
    negative = ndl[0, 0:n] <= mininf
    if negative.any() or not np.all(np.isfinite(ndl)):
        warn('Bad entries: ' + str(negative.nonzero()[0]), PydlutilsUserWarning)
        return negative.nonzero()[0], ndl
    try:
        lower = cholesky_banded(ndl[:, 0:n], lower=True)
    except LinAlgError:
        #
        # Figure out where the error is.
        #
        lower = ndl.copy()
        kn = bw - 1
        spot = np.arange(kn, dtype='i4') + 1
        for j in range(n):
            lower[0, j] = np.sqrt(lower[0, j])
            lower[spot, j] /= lower[0, j]
            x = lower[spot, j]
            if not np.all(np.isfinite(x)):
                warn('NaN found in cholesky_band.', PydlutilsUserWarning)
                return j, ndl
    #
    # Restore padding.
    #
    ll = np.zeros(ndl.shape, dtype=ndl.dtype)
    ll[:, 0:n] = lower
    return -1, ll


def cholesky_solve(a, bb):
    """Solve the equation :math:`A x = b` where `a` is a *lower*
    Cholesky-banded matrix.

    In the :class:`~pydl.pydlutils.bspline.Bspline` machinery, `a` needs to
    be padded.  This function should only used with the output of
    :func:`~pydl.pydlutils.bspline.cholesky_band`, to ensure the proper
    padding on `a`.  Otherwise the computation is delegated to
    :func:`scipy.linalg.cho_solve_banded`.

    Parameters
    ----------
    a : :class:`numpy.ndarray`
        *Lower* Cholesky decomposition of :math:`A` in :math:`A x = b`.
    bb : :class:`numpy.ndarray`
        :math:`b` in :math:`A x = b`.

    Returns
    -------
    :class:`numpy.ndarray`
        The solution, padded to be the same shape as `bb`.
    """
    bw = a.shape[0]
    n = bb.shape[0] - bw
    x = np.zeros(bb.shape, dtype=bb.dtype)
    x[0:n] = cho_solve_banded((a[:, 0:n], True), bb[0:n])
    return x


def iterfit(xdata, ydata, invvar=None, upper=5, lower=5, x2=None,
            maxiter=10, nord=4, bkpt=None, fullbkpt=None,
            kwargs_bspline={}, kwargs_reject={}):
    """Iteratively fit a B-spline set to data, with rejection.

    Parameters
    ----------
    xdata : :class:`numpy.ndarray`
        Independent variable.
    ydata : :class:`numpy.ndarray`
        Dependent variable.
    invvar : :class:`numpy.ndarray`, optional
        Inverse variance of `ydata`.  If not set, it will be calculated based
        on the standard deviation.
    upper : :class:`int` or :class:`float`, optional
        Upper rejection threshold in units of sigma, defaults to 5 sigma.
    lower : :class:`int` or :class:`float`, optional
        Lower rejection threshold in units of sigma, defaults to 5 sigma.
    x2 : :class:`numpy.ndarray`, optional
        Orthogonal dependent variable for 2d fits.
    maxiter : :class:`int`, optional
        Maximum number of rejection iterations, default 10.  Set this to
        zero to disable rejection.
    nord : :class:`int`
    bkpt : :None:
    fullbkpt : :class:`numpy.ndarray`
    kwargs_bspline : :class:`list`
    kwargs_reject : :class:`list`

    Returns
    -------
    :func:`tuple`
        A tuple containing the fitted bspline object and an output mask.
    """
    nx = xdata.size
    if ydata.size != nx:
        raise ValueError('Dimensions of xdata and ydata do not agree.')
    if invvar is not None:
        if invvar.size != nx:
            raise ValueError('Dimensions of xdata and invvar do not agree.')
    else:
        #
        # This correction to the variance makes it the same
        # as IDL's variance()
        #
        var = ydata.var()*(float(nx)/float(nx-1))
        if var == 0:
            var = 1.0
        invvar = np.ones(ydata.shape, dtype=ydata.dtype)/var
    if x2 is not None:
        if x2.size != nx:
            raise ValueError('Dimensions of xdata and x2 do not agree.')
    yfit = np.zeros(ydata.shape, dtype=ydata.dtype)
    if invvar.size == 1:
        outmask = True
    else:
        outmask = np.ones(invvar.shape, dtype='bool')
    xsort = xdata.argsort()
    maskwork = (outmask & (invvar > 0))[xsort]
    if 'oldset' in kwargs_bspline:
        sset = kwargs_bspline['oldset']
        sset.mask = True
        sset.coeff = 0
    else:
        if not maskwork.any():
            raise ValueError('No valid data points.')
#        if 'fullbkpt' in kwargs:
#            fullbkpt = kwargs['fullbkpt']
        else:
            sset = Bspline(xdata[xsort[maskwork]], nord=nord, bkpt=bkpt,
                           fullbkpt=fullbkpt, **kwargs_bspline)
            if maskwork.sum() < sset.nord:
                warn('Number of good data points fewer than nord.',
                     PydlutilsUserWarning)
                return sset, outmask
            if x2 is not None:
                if 'xmin' in kwargs_bspline:
                    xmin = kwargs_bspline['xmin']
                else:
                    xmin = x2.min()
                if 'xmax' in kwargs_bspline:
                    xmax = kwargs_bspline['xmax']
                else:
                    xmax = x2.max()
                if xmin == xmax:
                    xmax = xmin + 1
                sset.xmin = xmin
                sset.xmax = xmax
                if 'funcname' in kwargs_bspline:
                    sset.funcname = kwargs_bspline['funcname']
    xwork = xdata[xsort]
    ywork = ydata[xsort]
    invwork = invvar[xsort]
    if x2 is not None:
        x2work = x2[xsort]
    else:
        x2work = None
    iiter = 0
    error = 0
    qdone = False
    while (error != 0 or qdone is False) and iiter <= maxiter:
        goodbk = sset.mask.nonzero()[0]
        if maskwork.sum() <= 1 or not sset.mask.any():
            sset.coeff = 0
            iiter = maxiter + 1  # End iterations
        else:
            if 'requiren' in kwargs_bspline:
                i = 0
                while xwork[i] < sset.breakpoints[goodbk[sset.nord]] and i < nx-1:
                    i += 1
                ct = 0
                for ileft in range(sset.nord, sset.mask.sum()-sset.nord+1):
                    while (sset.breakpoints[goodbk[ileft]] <= xwork[i]
                           < sset.breakpoints[goodbk[ileft + 1]] and
                           i < nx - 1):
                        ct += invwork[i]*maskwork[i] > 0
                        i += 1
                    if ct >= kwargs_bspline['requiren']:
                        ct = 0
                    else:
                        sset.mask[goodbk[ileft]] = False
            error, yfit = sset.fit(xwork, ywork, invwork*maskwork,
                                   x2=x2work)
        iiter += 1
        inmask = maskwork
        if error == -2:

            return sset, outmask
        elif error == 0:
            maskwork, qdone = djs_reject(ywork, yfit, invvar=invwork,
                                         inmask=inmask, outmask=maskwork,
                                         upper=upper, lower=lower,
                                         **kwargs_reject)
        else:
            pass
    outmask[xsort] = maskwork
    temp = yfit
    yfit[xsort] = temp
    return sset, outmask



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_writer
from kcwidrp.core.bokeh_plotting import bokeh_plot
from kcwidrp.core.kcwi_plotting import save_plot

from bokeh.plotting import figure
import numpy as np
from scipy.signal import savgol_filter
import time


class SubtractScatteredLight(BasePrimitive):
    """
    Subtract scattered light between slices.

    Uses the centeral, un-illuminated part of the image to generate a model of
    the scattered light.  Subtract this model from the entire image.

    Uses the following configuration parameter:

        * skipscat: set to ``True`` to skip scattered light subtraction. Defaults to ``False``.

    Writes out a \*_intd.fits file regardless if scattered light is subtracted
    or not.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _perform(self):

        # Header keyword to update
        key = 'SCATSUB'
        keycom = "was scattered light subtracted?"

        # get camera
        camera = self.action.args.ccddata.header['CAMERA'].upper()

        # Skip if nod-and-shuffle
        if self.action.args.nasmask:
            self.logger.info("NAS Mask: skipping scattered light subtraction")
            self.action.args.ccddata.header[key] = (False, keycom)
        elif self.config.instrument.skipscat:
            self.logger.info("Skipping scattered light subtraction by request")
            self.action.args.ccddata.header[key] = (False, keycom)
        else:
            # Binning
            # ybin = self.action.args.ybinsize
            # Get size of image
            siz = self.action.args.ccddata.data.shape
            # Get x range for scattered light
            x0 = int(siz[1] / 2 - 180 / self.action.args.xbinsize)
            x1 = int(siz[1] / 2 + 180 / self.action.args.xbinsize)
            # Get y limits
            y0 = 0
            # y1 = int(siz[0] / 2 - 1)
            # y2 = y1 + 1
            y3 = siz[0]
            # print("x limits: %d, %d, y limits: %d, %d" % (x0, x1, y0, y3))
            # Y data values
            yvals = np.nanmedian(self.action.args.ccddata.data[y0:y3, x0:x1],
                                 axis=1)
            # Fix extreme edge values
            if 'RED' in camera:
                if np.nanmax(yvals) > 100:
                    yvals[:10] = yvals[10:20]
                    yvals[-10:] = yvals[-21:-11]
            else:
                yvals[0] = np.nanmedian(yvals[1:10])
                yvals[-1] = np.nanmedian(yvals[-11:-2])
            # X data values
            xvals = np.arange(len(yvals), dtype=np.float)
            # filter window
            fwin = 151
            scat = savgol_filter(yvals, fwin, 3)
            signal_to_noise = np.mean(scat) / np.nanstd(yvals - scat)
            if signal_to_noise < 25.:
                if signal_to_noise < 5:
                    fwin = 501
                else:
                    fwin = 303
                scat = savgol_filter(yvals, fwin, 3)
                signal_to_noise = np.mean(scat) / np.nanstd(yvals - scat)
            self.logger.info("Smoothing scattered light with window of %d px"
                             % fwin)
            self.logger.info("Mean signal to noise = %.2f" % signal_to_noise)
            if self.config.instrument.plot_level >= 1:
                # output filename stub
                scfnam = "scat_%05d_%s_%s_%s" % \
                         (self.action.args.ccddata.header['FRAMENO'],
                          self.action.args.illum, self.action.args.grating,
                          self.action.args.ifuname)
                # plot
                p = figure(title=self.action.args.plotlabel +
                           "SCATTERED LIGHT FIT",
                           x_axis_label='y pixel', y_axis_label='e-',
                           plot_width=self.config.instrument.plot_width,
                           plot_height=self.config.instrument.plot_height)
                p.circle(xvals, yvals, legend_label="Scat")
                p.line(xvals, scat, color='red', line_width=3,
                       legend_label="Smooth")
                p.legend.location = "top_left"
                bokeh_plot(p, self.context.bokeh_session)
                if self.config.instrument.plot_level >= 2:
                    input("Next? <cr>: ")
                else:
                    time.sleep(self.config.instrument.plot_pause)
                save_plot(p, filename=scfnam+".png")
            # Subtract scattered light
            self.logger.info("Starting scattered light subtraction")
            for ix in range(0, siz[1]):
                self.action.args.ccddata.data[y0:y3, ix] = \
                    self.action.args.ccddata.data[y0:y3, ix] - scat
            self.action.args.ccddata.header[key] = (True, keycom)

        log_string = SubtractScatteredLight.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.logger.info(log_string)

        # write out int image
        kcwi_fits_writer(self.action.args.ccddata,
                         table=self.action.args.table,
                         output_file=self.action.args.name,
                         output_dir=self.config.instrument.output_directory,
                         suffix="intd")
        self.context.proctab.update_proctab(frame=self.action.args.ccddata,
                                            suffix="intd", 
                                            filename=self.action.args.name)
        self.context.proctab.write_proctab(tfil=self.config.instrument.procfile)

        return self.action.args
    # END: SubtractScatteredLight()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.core.bokeh_plotting import bokeh_plot
from kcwidrp.primitives.kcwi_file_primitives import plotlabel

import numpy as np
import logging
from bokeh.util.logconfig import basicConfig
from bokeh.plotting import figure
from scipy.signal import find_peaks
import time


class FindBars(BasePrimitive):
    """
    Find bars in middle row of cont bars image

    Uses the middle row with surrounding rows to find the individual continuum
    bar locations in the image.  Checks to be sure the correct number of bars
    has been found (see NBARS in kcwidrp/configs/kcwi.cfg).

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger
        basicConfig(level=logging.ERROR)

    def _pre_condition(self):
        self.logger.info("Checking for master contbars")
        if 'MCBARS' in self.action.args.ccddata.header['IMTYPE']:
            return True
        else:
            return False

    def _perform(self):
        self.logger.info("Finding continuum bars")
        # initialize
        reference_bar = self.config.instrument.REFBAR
        middle_centers = []
        # get image dimensions
        x_size = self.action.args.ccddata.data.shape[1]
        y_size = self.action.args.ccddata.data.shape[0]
        # get binning
        y_binning = self.action.args.ybinsize
        # get camera
        camera = self.action.args.camera
        # get grating
        grating = self.action.args.grating
        window = int(10 / y_binning)
        # get plot label
        plab = plotlabel(self.action.args)
        # select from center rows of image
        div_fac = 2
        middle_y_row = int(y_size / div_fac)
        # handle dichroic edge by moving middle until we get enough peaks
        n_tries = 0
        peaks_found = 0
        middle_vector = None
        bar_thresh = None
        average_value_middle_vector = None
        stdev_value_middle_vector = None
        peaks_vector = None
        while peaks_found != self.config.instrument.NBARS and n_tries < 5:
            self.logger.info("Middle row for bars finding: %d" % middle_y_row)
            middle_vector = np.median(
                self.action.args.ccddata.data[
                    (middle_y_row-window):(middle_y_row+window+1), :], axis=0)
            # set threshold for peak finding
            average_value_middle_vector = np.average(middle_vector)
            stdev_value_middle_vector = np.nanstd(middle_vector)
            bar_thresh = average_value_middle_vector + \
                0.5 * stdev_value_middle_vector
            self.logger.info("peak threshold = %f" % bar_thresh)
            # find peaks above threshold
            peaks_in_middle_vector, _ = find_peaks(
                middle_vector, height=bar_thresh)
            peaks_vector = []
            for pk in peaks_in_middle_vector:
                if 5 < pk < (self.action.args.ccddata.header['NAXIS1'] - 5):
                    peaks_vector.append(pk)
            peaks_found = len(peaks_vector)
            if self.config.instrument.plot_level >= 3:
                # plot the peak positions
                x = np.arange(len(middle_vector))
                p = figure(
                    title=plab + "BARS TRACE at = %d, TRY %d" % (middle_y_row,
                                                                 n_tries),
                    x_axis_label='CCD X (px)', y_axis_label='e-',
                    plot_width=self.config.instrument.plot_width,
                    plot_height=self.config.instrument.plot_height)
                p.line(x, middle_vector, color='blue', legend_label="Trace")
                p.scatter(peaks_vector,
                          middle_vector[peaks_vector], marker='x',
                          color='red', legend_label="FoundBar")
                p.line([0, x_size], [bar_thresh, bar_thresh],
                       color='grey', line_dash='dashed')
                p.legend.location = "bottom_center"
                bokeh_plot(p, self.context.bokeh_session)
                input("Next? <cr>: ")

            n_tries += 1
            # do we have the requisite number?
            if peaks_found != self.config.instrument.NBARS:
                self.logger.warning("Did not find %d peaks: n peaks = %d" %
                                    (self.config.instrument.NBARS, peaks_found))
                self.logger.info("Calculating new middle row and retrying")
                if camera == 0:     # BLUE
                    div_fac += 0.5
                else:               # RED
                    if 'RH4' in grating:
                        div_fac += 0.5
                    else:
                        div_fac -= 0.5
                middle_y_row = int(y_size / div_fac)
        if peaks_found != self.config.instrument.NBARS:
            self.logger.error("Did not find %d peaks: n peaks = %d" %
                              (self.config.instrument.NBARS, peaks_found))
        else:
            self.logger.info("found %d bars" % peaks_found)

            if self.config.instrument.plot_level >= 1:
                # plot the peak positions
                x = np.arange(len(middle_vector))
                p = figure(
                    title=plab + "BARS MID TRACE Thresh = %.2f" % bar_thresh,
                    x_axis_label='CCD X (px)', y_axis_label='e-',
                    plot_width=self.config.instrument.plot_width,
                    plot_height=self.config.instrument.plot_height)
                p.line(x, middle_vector, color='blue', legend_label="MidTrace")
                p.scatter(peaks_vector,
                          middle_vector[peaks_vector], marker='x',
                          color='red', legend_label="FoundBar")
                p.line([0, x_size], [bar_thresh, bar_thresh],
                       color='grey', line_dash='dashed')
                p.legend.location = "bottom_center"
                bokeh_plot(p, self.context.bokeh_session)
                if self.config.instrument.plot_level >= 2:
                    input("Next? <cr>: ")
                else:
                    time.sleep(self.config.instrument.plot_pause)
                # calculate the bar centroids

            # Do we plot?
            if self.config.instrument.plot_level >= 3:
                do_inter = True
            else:
                do_inter = False
            for ip, peak in enumerate(peaks_vector):
                xs = list(range(peak-window, peak+window+1))
                ys = middle_vector[xs] - np.nanmin(middle_vector[xs])
                xc = np.sum(xs*ys) / np.sum(ys)
                middle_centers.append(xc)
                if do_inter:
                    p = figure(
                        title=plab + "BAR %d CENTRD = %.2f" % (ip, xc),
                        x_axis_label='CCD X (px)', y_axis_label='e-',
                        plot_width=self.config.instrument.plot_width,
                        plot_height=self.config.instrument.plot_height)
                    p.line(xs, ys, color='blue', legend_label='Bar Trace')
                    p.circle(xs, ys, color='red', legend_label='Bar Trace')
                    p.line([xc, xc], [bar_thresh, middle_vector[peak]],
                           color='green', legend_label='Cntrd')
                    bokeh_plot(p, self.context.bokeh_session)
                    if do_inter:
                        q = input("Next? <cr>, q - quit: ")
                        if 'Q' in q.upper():
                            do_inter = False
                    else:
                        time.sleep(self.config.instrument.plot_pause)
            self.logger.info("Found middle centroids for continuum bars")
        # store threshold
        self.action.args.bar_thresh = bar_thresh
        self.action.args.bar_avg = average_value_middle_vector
        self.action.args.bar_std = stdev_value_middle_vector
        # store peaks
        self.action.args.middle_centers = middle_centers
        # store the row where we got them
        self.action.args.middle_row = middle_y_row
        self.action.args.window = window
        # calculate reference delta x based on refbar
        self.action.args.reference_delta_x = 0.
        try:
            if (reference_bar-1) > 0 and \
                    (reference_bar+3) < self.config.instrument.NBARS:
                for ib in range(reference_bar-1, reference_bar+3):
                    self.action.args.reference_delta_x += \
                        (middle_centers[ib] - middle_centers[ib-1])
                ndiv = 4.
            else:
                for ib in range(reference_bar, reference_bar+1):
                    self.action.args.reference_delta_x += \
                        (middle_centers[ib] - middle_centers[ib-1])
                ndiv = 2.
        except IndexError:
            self.logger.warning(
                "Not enough bars per slice to determine ref x sep")
            self.action.args.reference_delta_x = 1.
            ndiv = 1.
        self.action.args.reference_delta_x /= ndiv
        # store image info
        self.action.args.contbar_image_number = \
            self.action.args.ccddata.header['FRAMENO']
        self.action.args.contbar_image = \
            self.action.args.ccddata.header['OFNAME']

        log_string = FindBars.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.logger.info(log_string)

        return self.action.args
    # END: class FindBars()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import strip_fname, plotlabel
from kcwidrp.core.bokeh_plotting import bokeh_plot
from kcwidrp.core.kcwi_plotting import set_plot_lims, save_plot
import os
import numpy as np
from scipy.interpolate import interpolate
from bokeh.plotting import figure
from astropy.io import fits as pf


class SolveAIT(BasePrimitive):
    """
    Solve the AIT data of the IFU.

    This routine was only used during KCRM AIT at Caltech when we only had three
    bars to work with.  It is here only in case there is a need to reprocess the
    AIT data.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.action.args.geometry_file = None
        self.action.args.x0out = None
        self.action.args.wave0out = None
        self.action.args.wave1out = None
        self.action.args.wavegood0 = None
        self.action.args.wavegood1 = None
        self.action.args.waveall0 = None
        self.action.args.waveall1 = None
        self.action.args.wavemid = None
        self.logger = context.pipeline_logger

    def _pre_condition(self):
        self.logger.info("Checking for master arc")
        if 'MARC' in self.action.args.ccddata.header['IMTYPE']:
            return True
        else:
            return False

    def _perform(self):
        self.logger.info("Writing out AIT spectra")

        log_string = SolveAIT.__module__

        do_plot = (self.config.instrument.plot_level >= 1)

        # Get some geometry constraints
        goody0 = 0
        goody1 = max(self.action.args.xsvals)
        # N&S limits
        goodnsy0 = self.action.args.shufrows
        goodnsy1 = goodnsy0 + self.action.args.shufrows
        # Calculate wavelength ranges
        y0wvs = []
        y1wvs = []
        # N&S ranges
        y0nswvs = []
        y1nswvs = []
        # Get wavelength extremes for each bar
        for fcfs in self.action.args.fincoeff:
            y0wvs.append(float(np.polyval(fcfs, goody0)))
            y1wvs.append(float(np.polyval(fcfs, goody1)))
            y0nswvs.append(float(np.polyval(fcfs, goodnsy0)))
            y1nswvs.append(float(np.polyval(fcfs, goodnsy1)))
        # Now get ensemble extremes
        y0max = max(y0wvs)
        y0min = min(y0wvs)
        y1max = max(y1wvs)
        y1min = min(y1wvs)
        y0nsmax = max(y0nswvs)
        y0nsmin = min(y0nswvs)
        y1nsmax = max(y1nswvs)
        y1nsmin = min(y1nswvs)
        # Cube trimming wavelengths
        trimw0 = y0min
        trimw1 = y1max
        # Check for negative dispersion
        if trimw0 > trimw1:
            trimw0 = y1min
            trimw1 = y0max
        # Calculate output wavelengths
        dwout = self.action.args.dwout
        self.logger.info("Output Delta WAVE: %.3f", dwout)
        ndels = int((trimw0 - self.config.instrument.WAVEFID) / dwout)
        self.action.args.wave0out = \
            self.config.instrument.WAVEFID + float(ndels) * dwout
        ndels = int((trimw1 - self.config.instrument.WAVEFID) / dwout)
        self.action.args.wave1out = \
            self.config.instrument.WAVEFID + float(ndels) * dwout
        self.logger.info("WAVE RANGE: %.2f - %.2f" %
                         (self.action.args.wave0out, self.action.args.wave1out))
        # Calculate wavelength limits
        self.action.args.wavegood0 = min([y0max, y1max])
        self.action.args.wavegood1 = max([y0min, y1min])
        self.action.args.waveall0 = min([y0min, y1min])
        self.action.args.waveall1 = max([y0max, y1max])
        self.action.args.wavemid = np.average([self.action.args.wavegood0,
                                               self.action.args.wavegood1,
                                               self.action.args.waveall0,
                                               self.action.args.waveall1])
        self.action.args.wavensgood0 = min([y0nsmax, y1nsmax])
        self.action.args.wavensgood1 = max([y0nsmin, y1nsmin])
        self.action.args.wavensall0 = min([y0nsmin, y1nsmin])
        self.action.args.wavensall1 = max([y0nsmax, y1nsmax])
        self.action.args.wavensmid = np.average([self.action.args.wavensgood0,
                                                 self.action.args.wavensgood1,
                                                 self.action.args.wavensall0,
                                                 self.action.args.wavensall1])
        self.logger.info("WAVE  GOOD: %.2f - %.2f" %
                         (self.action.args.wavegood0,
                          self.action.args.wavegood1))
        self.logger.info("WAVE   ALL: %.2f - %.2f" %
                         (self.action.args.waveall0, self.action.args.waveall1))
        self.logger.info("WAVE   MID: %.2f" % self.action.args.wavemid)

        # Use extremes to define output size
        ysize = int((self.action.args.waveall1 - self.action.args.wave0out)
                    / dwout)
        yout = np.arange(ysize) * dwout + self.action.args.wave0out
        xsize = self.config.instrument.NBARS * 2 + 1
        self.logger.info("Output spectra will be %d x %d px" % (xsize, ysize))
        specout = np.zeros((ysize, xsize), dtype=float)
        # Loop over AIT bar traces
        for ib in range(0, self.config.instrument.NBARS):
            fc = self.action.args.fincoeff[ib]
            arc = self.context.arcs[ib]
            bar = self.context.bars[ib]
            xwv = np.polyval(fc, np.arange(len(arc)))
            arcint = interpolate.interp1d(xwv, arc, bounds_error=False,
                                          fill_value='extrapolate')
            barint = interpolate.interp1d(xwv, bar, bounds_error=False,
                                          fill_value='extrapolate')
            arcout = arcint(yout)
            barout = barint(yout)
            specout[:, ib] = arcout[:]
            specout[:, ib + self.config.instrument.NBARS] = barout[:]
            if do_plot:
                # plot output name stub
                pfname = "arc_%05d_%s_%s_%s_tf%02d" % (
                    self.action.args.ccddata.header['FRAMENO'],
                    self.action.args.illum, self.action.args.grating,
                    self.action.args.ifuname,
                    int(100 * self.config.instrument.TAPERFRAC))
                wall0 = self.action.args.waveall0
                wall1 = self.action.args.waveall1
                wgoo0 = self.action.args.wavegood0
                wgoo1 = self.action.args.wavegood1
                wl_good = [i for i, v in enumerate(yout) if wgoo0 <= v <= wgoo1]
                yran = [np.nanmin([np.nanmin(arcout[wl_good]),
                                   np.nanmin(barout[wl_good])]),
                        np.nanmax([np.nanmax(arcout[wl_good]),
                                   np.nanmax(barout[wl_good])])]
                p = figure(title=plotlabel(self.action.args) + "ARC # %d" % ib,
                           x_axis_label="Wavelength",
                           y_axis_label="Flux",
                           plot_width=self.config.instrument.plot_width,
                           plot_height=self.config.instrument.plot_height)
                p.line(yout, arcout, legend_label='Arc', color='blue')
                p.line(yout, barout, legend_label='Bar', color='red')
                p.line([wgoo0, wgoo0], yran, legend_label='Good WLs',
                       color='green')
                p.line([wgoo1, wgoo1], yran, color='green')
                p.line([wall0, wall0], yran, legend_label='All WLs',
                       color='orange')
                p.line([wall1, wall1], yran, color='orange')
                set_plot_lims(p, xlim=[wall0, wall1], ylim=yran)
                bokeh_plot(p, self.context.bokeh_session)
                save_plot(p, filename=pfname + '_bar%d.png' % ib)
                if self.config.instrument.plot_level >= 2:
                    _ = input("Next? <cr>: ")
        specout[:, -1] = yout[:]

        # Pixel scales
        pxscl = self.config.instrument.PIXSCALE * self.action.args.xbinsize
        ifunum = self.action.args.ifunum
        if ifunum == 2:
            slscl = self.config.instrument.SLICESCALE / 2.0
        elif ifunum == 3:
            slscl = self.config.instrument.SLICESCALE / 4.0
        else:
            slscl = self.config.instrument.SLICESCALE

        # Update header
        # Geometry corrected?
        self.action.args.ccddata.header['GEOMCOR'] = (
            True, 'Geometry corrected?')
        #
        # Spatial geometry
        # self.action.args.ccddata.header['BARSEP'] = (
        #    self.action.args.barsep, 'separation of bars (binned pix)')
        # self.action.args.ccddata.header['BAR0'] = (
        #    self.action.args.bar0, 'first bar pixel position')
        # Wavelength ranges
        if self.action.args.nasmask:
            self.action.args.ccddata.header['WAVALL0'] = (
                self.action.args.wavensall0, 'Low inclusive wavelength')
            self.action.args.ccddata.header['WAVALL1'] = (
                self.action.args.wavensall1, 'High inclusive wavelength')
            self.action.args.ccddata.header['WAVGOOD0'] = (
                self.action.args.wavensgood0, 'Low good wavelength')
            self.action.args.ccddata.header['WAVGOOD1'] = (
                self.action.args.wavensgood1, 'High good wavelength')
            self.action.args.ccddata.header['WAVMID'] = (
                self.action.args.wavensmid, 'middle wavelength')
        else:
            self.action.args.ccddata.header['WAVALL0'] = (
                self.action.args.waveall0, 'Low inclusive wavelength')
            self.action.args.ccddata.header['WAVALL1'] = (
                self.action.args.waveall1, 'High inclusive wavelength')
            self.action.args.ccddata.header['WAVGOOD0'] = (
                self.action.args.wavegood0, 'Low good wavelength')
            self.action.args.ccddata.header['WAVGOOD1'] = (
                self.action.args.wavegood1, 'High good wavelength')
            self.action.args.ccddata.header['WAVMID'] = (
                self.action.args.wavemid, 'middle wavelength')
        # Dichroic fraction
        try:
            dichroic_fraction = self.action.args.dich_frac
        except AttributeError:
            dichroic_fraction = 1.
        self.action.args.ccddata.header['DICHFRAC'] = (
            dichroic_fraction, 'Dichroic Fraction')
        # Wavelength fit statistics
        self.action.args.ccddata.header['AVWVSIG'] = (
            self.action.args.av_bar_sig, 'Avg. bar wave sigma (Ang)')
        self.action.args.ccddata.header['SDWVSIG'] = (
            self.action.args.st_bar_sig, 'Stdev. var wave sigma (Ang)')
        # Pixel scales
        self.action.args.ccddata.header['PXSCL'] = (
            pxscl, 'Pixel scale along slice (deg)')
        self.action.args.ccddata.header['SLSCL'] = (
            slscl, 'Pixel scale perp. to slices (deg)')
        # Geometry origins
        self.action.args.ccddata.header['CBARSNO'] = (
            self.action.args.contbar_image_number,
            'Continuum bars image number')
        self.action.args.ccddata.header['CBARSFL'] = (
            self.action.args.contbar_image, 'Continuum bars image filename')
        self.action.args.ccddata.header['ARCNO'] = (
            self.action.args.arc_number, 'Arc image number')
        self.action.args.ccddata.header['ARCFL'] = (
            self.action.args.arc_image, 'Arc image filename')
        # self.action.args.ccddata.header['GEOMFL'] = (
        #     geom_file.split('/')[-1], 'Geometry file')
        # WCS
        # self.action.args.ccddata.header['IFUPA'] = (
        #     skypa, 'IFU position angle (degrees)')
        # self.action.args.ccddata.header['IFUROFF'] = (
        #    self.config.instrument.ROTOFF, 'IFU-SKYPA offset (degrees)')
        self.action.args.ccddata.header['WCSDIM'] = (
            2, 'number of dimensions in WCS')
        self.action.args.ccddata.header['WCSNAME'] = 'KCRM_AIT'
        self.action.args.ccddata.header['CTYPE1'] = 'Pix'
        self.action.args.ccddata.header['CTYPE2'] = ('AWAV',
                                                     'Air Wavelengths')
        self.action.args.ccddata.header['CUNIT1'] = ('pixel', 'image units')
        self.action.args.ccddata.header['CUNIT2'] = ('Angstrom',
                                                     'Wavelength units')
        self.action.args.ccddata.header['CNAME1'] = ('KCWI Pix', 'pix name')
        self.action.args.ccddata.header['CNAME2'] = ('KCWI Wavelength',
                                                     'Wavelength name')
        self.action.args.ccddata.header['CRVAL1'] = (1, 'Pixel zeropoint')
        self.action.args.ccddata.header['CRVAL2'] = (self.action.args.wave0out,
                                                     'Wavelength zeropoint')
        self.action.args.ccddata.header['CDELT1'] = (1.0, 'Pixel scale')
        self.action.args.ccddata.header['CDELT2'] = (dwout, 'Wavelength scale')
        self.action.args.ccddata.header['CRPIX1'] = (1,
                                                     'Pixel reference pixel')
        self.action.args.ccddata.header['CRPIX2'] = (
            1, 'Wavelength reference pixel')

        # write out cube
        self.action.args.ccddata.header['HISTORY'] = log_string

        # Package AIT spectra
        ofname = self.action.args.ccddata.header['OFNAME']
        self.action.args.ait_file = os.path.join(
            self.config.instrument.output_directory,
            strip_fname(ofname) + '_aitspec.fits')
        if os.path.exists(self.action.args.ait_file):
            self.logger.error("AIT spec file already exists: %s" %
                              self.action.args.ait_file)
        else:

            hdu = pf.PrimaryHDU(specout, header=self.action.args.ccddata.header)
            hdul = pf.HDUList([hdu])
            hdul.writeto(self.action.args.ait_file)

            self.logger.info("AIT spectra written to: %s" %
                             self.action.args.ait_file)

        self.logger.info(log_string)

        return self.action.args
    # END: class SolveAIT()



import os
import requests
import time

from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_writer, \
                                                    kcwi_fits_reader, \
                                                    strip_fname


class SendHTTP(BasePrimitive):

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger
    
    def _pre_condition(self):
        self.user = self.config.rti.rti_user
        self.pw = self.config.rti.rti_pass
        if self.user == '' or self.pw == '':
            self.logger.error("Username or password is not set for RTI access")
            return False
        return True

    def _perform(self):

        if not self.action.args.ccddata.header['KOAID']:
            self.logger.error(f"Encountered a file with no KOA ID: {self.action.args.name}")
            return self.action.args
        
        data_directory = os.path.join(self.config.rti.cwd,
                                      self.config.rti.output_directory)
        
        self.logger.info(f"Alerting RTI that {strip_fname(self.action.args.name)} is ready for ingestion")

        url = self.config.rti.rti_url
        data = {
            'instrument': 'KCWI',
            'koaid': self.action.args.ccddata.header['KOAID'],
            'ingesttype': self.config.rti.rti_ingesttype,
            'datadir': str(data_directory),
            'start': str(self.action.args.ingest_time),
            'reingest': self.config.rti.rti_reingest,
            'testonly': self.config.rti.rti_testonly,
            'dev': self.config.rti.rti_dev
        }
        
        attempts = 0
        limit = self.config.rti.rti_attempts
        while attempts < limit:
            res = self.get_url(url, data)
            if res is None:
                t = self.config.rti.rti_retry_time
                attempts += 1
                self.logger.error(f"Waiting {t} seconds to attempt again... ({attempts}/{limit})")
                time.sleep(t)
            else:
                self.logger.info(f"Post returned status code {res.status_code}")
                return self.action.args
        
        self.logger.error(f"Post attempted {limit} times and got no response.")
        self.logger.error("Aborting.")
        return self.action.args
    
    def get_url(self, url, data):
        try:
            res = requests.get(url, params = data, auth=(
                                                        self.user,
                                                        self.pw
                                                        ))
            self.logger.info(f"Sending {res.request.url}")
        except requests.exceptions.RequestException as e:
            self.logger.error(f"Error caught while posting to {url}:")
            self.logger.error(e)
            return None
        return res



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.core.bokeh_plotting import bokeh_plot
from kcwidrp.core.kcwi_plotting import save_plot
from kcwidrp.core.bokeh_plotting import bokeh_clear

from bokeh.plotting import figure
from bokeh.layouts import gridplot
import numpy as np
import math
import time


class SubtractOverscan(BasePrimitive):
    """
    Determines overscan offset and subtracts it from the image.

    Uses the BIASSEC header keyword to determine where to calculate the overscan
    offset.  Subtracts the overscan offset and records the value in the header.
    In addition, performs a polynomial fit and uses the residuals to determine
    the read noise in the overscan.  Records the overscan readnoise in the
    header as well.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _perform(self):
        # image sections for each amp
        bsec, dsec, tsec, direc, amps, aoff = self.action.args.map_ccd
        namps = len(amps)
        # polynomial fit order
        if namps == 4:
            porder = 2
        else:
            porder = 7
        minoscanpix = self.config.instrument.minoscanpix
        oscanbuf = self.config.instrument.oscanbuf
        frameno = self.action.args.ccddata.header['FRAMENO']
        # header keyword to update
        key = 'OSCANSUB'
        keycom = 'Overscan subtracted?'
        # is it performed?
        performed = False
        # loop over amps
        plts = []   # plots for each amp

        for ia in amps:
            # bias correct amp number for indexing python arrays
            iac = ia - aoff
            # get gain
            gain = self.action.args.ccddata.header['GAIN%d' % ia]
            # check if we have enough data to fit
            if (bsec[iac][3] - bsec[iac][2]) > minoscanpix:
                # pull out an overscan vector
                x0 = bsec[iac][2] + oscanbuf
                x1 = bsec[iac][3] - oscanbuf
                y0 = bsec[iac][0]
                y1 = bsec[iac][1] + 1
                # get overscan value to subtract
                osval = int(np.nanmedian(
                    self.action.args.ccddata.data[y0:y1, x0:x1]))
                # vector to fit for determining read noise
                osvec = np.nanmedian(
                    self.action.args.ccddata.data[y0:y1, x0:x1], axis=1)
                nsam = x1 - x0
                xx = np.arange(len(osvec), dtype=np.float)
                # fit it, avoiding first 50 px
                if direc[iac][0]:
                    # forward read skips first 50 px
                    oscoef = np.polyfit(xx[50:], osvec[50:], porder)
                    # generate fitted overscan vector for full range
                    osfit = np.polyval(oscoef, xx)
                    # calculate residuals
                    resid = (osvec[50:] - osfit[50:]) * math.sqrt(nsam) * \
                        gain / 1.414
                else:
                    # reverse read skips last 50 px
                    oscoef = np.polyfit(xx[:-50], osvec[:-50], porder)
                    # generate fitted overscan vector for full range
                    osfit = np.polyval(oscoef, xx)
                    # calculate residuals
                    resid = (osvec[:-50] - osfit[:-50]) * math.sqrt(nsam) * \
                        gain / 1.414

                sdrs = float("%.3f" % np.std(resid))
                self.logger.info("Img # %05d, Amp %d [%d:%d, %d:%d]" % (frameno,
                                                                        ia,
                                                                        x0, x1,
                                                                        y0, y1))
                self.logger.info("Amp%d oscan counts (DN): %d" %
                                 (ia, osval))
                self.logger.info("Amp%d Read noise from oscan in e-: %.3f" %
                                 (ia, sdrs))
                self.action.args.ccddata.header['OSCNRN%d' % ia] = \
                    (sdrs, "amp%d RN in e- from oscan" % ia)
                self.action.args.ccddata.header['OSCNVAL%d' % ia] = \
                    (osval, "amp%d oscan counts (DN)" % ia)

                if self.config.instrument.plot_level >= 2:
                    x = np.arange(len(osvec))
                    p = figure(title='Img # %05d OSCAN [%d:%d, %d:%d] '
                                     'amp %d, noise: %.3f e-/px' %
                                     (frameno, x0, x1, y0, y1, ia, sdrs),
                               x_axis_label='overscan px',
                               y_axis_label='counts',
                               plot_width=self.config.instrument.plot_width,
                               plot_height=self.config.instrument.plot_height)
                    p.circle(x, osvec, legend_label="Data")
                    p.line(x, osfit, line_color='red', line_width=3,
                           legend_label="Fit")
                    bokeh_plot(p, self.context.bokeh_session)
                    plts.append(p)
                    if self.config.instrument.plot_level >= 3:
                        input("Next? <cr>: ")
                    else:
                        time.sleep(self.config.instrument.plot_pause)
                # subtract osval
                xx0 = dsec[iac][2]
                xx1 = dsec[iac][3] + 1
                self.action.args.ccddata.data[y0:y1, xx0:xx1] -= osval
                # for ix in range(dsec[iac][2], dsec[iac][3] + 1):
                #     self.action.args.ccddata.data[y0:y1, ix] = \
                #         self.action.args.ccddata.data[y0:y1, ix] - osfit
                performed = True
            else:
                self.logger.info("not enough overscan px to fit amp %d" % ia)
                performed = False
        if self.config.instrument.plot_level >= 3 and len(plts) > 0:
            bokeh_plot(gridplot(plts, ncols=(2 if namps > 2 else 1),
                                plot_width=500, plot_height=300,
                                toolbar_location=None),
                       self.context.bokeh_session)
            save_plot(gridplot(plts, ncols=(2 if namps > 2 else 1),
                               plot_width=500, plot_height=300,
                               toolbar_location=None),
                      filename="oscan_%05d.png" % frameno)
            if self.config.instrument.plot_level >= 2:
                input("Next? <cr>: ")
            else:
                time.sleep(self.config.instrument.plot_pause)
            bokeh_clear(self.context.bokeh_session)

        self.action.args.ccddata.header[key] = (performed, keycom)

        log_string = SubtractOverscan.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.logger.info(log_string)

        return self.action.args
    # END: class SubtractOverscan()



"""
Reduce KCWI data using the KCWI DRP running on the KeckDRPFramework

@author: skwok
"""

from keckdrpframework.core.framework import Framework
from keckdrpframework.config.framework_config import ConfigClass
from keckdrpframework.models.arguments import Arguments
from keckdrpframework.utils.drpf_logger import getLogger
from kcwidrp.core.bokeh_plotting import check_running_process
from kcwidrp.core.kcwi_get_std import is_file_kcwi_std

import subprocess
import time
import datetime
import argparse
import sys
import traceback
import os
import pkg_resources
import psutil
import shutil

from kcwidrp.pipelines.kcwi_pipeline import Kcwi_pipeline
from kcwidrp.core.kcwi_proctab import Proctab
import logging.config


def _parse_arguments(in_args: list) -> argparse.Namespace:
    description = "KCWI pipeline CLI"

    # this is a simple case where we provide a frame and a configuration file
    parser = argparse.ArgumentParser(prog=f"{in_args[0]}",
                                     description=description)
    parser.add_argument('-c', '--config', dest="kcwi_config_file", type=str,
                        help="KCWI configuration file", default=None)
    parser.add_argument('--write_config', dest="write_config",
                        help="Write out an editable config file in current dir"
                        " (kcwi.cfg)", action="store_true", default=False)
    parser.add_argument('-f', '--frames', nargs='*', type=str,
                        help='input image files (full path, list ok)',
                        default=None)
    parser.add_argument('-l', '--list', dest='file_list',
                        help='File containing a list of files to be processed',
                        default=None)
    parser.add_argument('-g', '--groups', dest='group_mode',
                        help='Use group mode: separate files by image type and '
                             'reduce in the correct order',
                        default=False, action='store_true')
    parser.add_argument('-t', '--taperfrac', dest='taperfrac', type=float,
                        help='Taper fraction for wavelength fitting',
                        default=None)
    parser.add_argument('-a', '--atlas_line_list', dest='atlas_line_list',
                        type=str, help="Atlas line list file", default=None)
    parser.add_argument('-M', '--middle_fraction', dest='middle_fraction',
                        type=float, help="Fraction of middle to use",
                        default=None)
    parser.add_argument('-o', '--atlas_offset', dest='atlas_offset',
                        type=int, help="Atlas offset (px)", default=None)
    parser.add_argument('-e', '--line_thresh', dest='line_thresh',
                        type=float, help="Line Cleaning Threshold (e-)",
                        default=None)
    parser.add_argument('-u', '--tukey_alpha', dest='tukey_alpha',
                        type=float, help="Tukey Window Alpha (0.0 - 1.0)",
                        default=None)
    parser.add_argument('-F', '--max_frac', dest='max_frac',
                        type=float, default=None,
                        help="Fraction of line max for fitting window "
                             "(default: 0.5)")

    # in this case, we are loading an entire directory,
    # and ingesting all the files in that directory
    parser.add_argument('-i', '--infiles', dest="infiles",
                        help="Input files, or pattern to match", nargs="?")
    parser.add_argument('-d', '--directory', dest="dirname", type=str,
                        help="Input directory", nargs='?', default=None)
    # after ingesting the files,
    # do we want to continue monitoring the directory?
    parser.add_argument('-m', '--monitor', dest="monitor",
                        help='Continue monitoring the directory '
                             'after the initial ingestion',
                        action='store_true', default=False)

    # special arguments, ignore
    parser.add_argument("-I", "--ingest_data_only", dest="ingest_data_only",
                        action="store_true",
                        help="Ingest data and terminate")
    parser.add_argument("-w", "--wait_for_event", dest="wait_for_event",
                        action="store_true", help="Wait for events")
    parser.add_argument("-W", "--continue", dest="continuous",
                        action="store_true",
                        help="Continue processing, wait for ever")
    parser.add_argument("-s", "--start_queue_manager_only",
                        dest="queue_manager_only", action="store_true",
                        help="Starts queue manager only, no processing",)

    # kcwi specific parameters
    parser.add_argument("-p", "--proctab", dest='proctab', help='Proctab file',
                        default=None)
    parser.add_argument("-b", "--blue", dest='blue', action="store_true",
                        default=False, help="KCWI Blue processing")
    parser.add_argument("-r", "--red", dest='red', action="store_true",
                        default=False, help="KCWI Red processing")
    parser.add_argument("-k", "--skipsky", dest='skipsky', action="store_true",
                        default=False, help="Skip sky subtraction")

    out_args = parser.parse_args(in_args[1:])
    return out_args


def check_directory(directory):
    if not os.path.isdir(directory):
        os.makedirs(directory)
        print("Directory %s has been created" % directory)


def main():

    # Package
    pkg = 'kcwidrp'

    # get arguments
    args = _parse_arguments(sys.argv)

    if args.write_config:
        dest = os.path.join(os.getcwd(), 'kcwi.cfg')
        if os.path.exists(dest):
            print("Config file kcwi.cfg already exists in current dir")
        else:
            kcwi_config_file = 'configs/kcwi.cfg'
            kcwi_config_fullpath = pkg_resources.resource_filename(
                pkg, kcwi_config_file)
            shutil.copy(kcwi_config_fullpath, os.getcwd())
            print("Copied kcwi.cfg into current dir.  Edit and use with -c")
        sys.exit(0)

    # This check can be removed once reduce_kcwi processes are
    # siloed against each other
    plist = []
    for p in psutil.process_iter():
        try:
            if "reduce_kcwi" in p.name():
                plist.append(p)
        except psutil.NoSuchProcess:
            continue
    # plist = [p for p in psutil.process_iter() if "reduce_kcwi" in p.name()]
    if len(plist) > 1:
        print("DRP already running in another process, exiting")
        sys.exit(0)

    def process_subset(in_subset):
        for in_frame in in_subset.index:
            arguments = Arguments(name=in_frame)
            framework.append_event('next_file', arguments, recurrent=True)
    
    def process_list(in_list):
        for in_frame in in_list:
            arguments = Arguments(name=in_frame)
            framework.append_event('next_file', arguments, recurrent=True)

    # make sure user has selected a channel
    if not args.blue and not args.red:
        print("\nERROR - DRP can process only one channel at a time\n\n"
              "Please indicate a channel to process:\n"
              "Either BLUE with -b or --blue or\n"
              "       RED  with -r or --red\n")
        sys.exit(0)

    if args.file_list:
        if '.fits' in args.file_list:
            print("\nERROR - trying to read in fits file as file list\n\n"
                  "Please use -f or --frames for direct input of fits files\n")
            sys.exit(0)

    # START HANDLING OF CONFIGURATION FILES ##########

    # check for the logs diretory
    check_directory("logs")
    # check for the plots directory
    check_directory("plots")

    framework_config_file = "configs/framework.cfg"
    framework_config_fullpath = \
        pkg_resources.resource_filename(pkg, framework_config_file)

    framework_logcfg_file = 'configs/logger.cfg'
    framework_logcfg_fullpath = \
        pkg_resources.resource_filename(pkg, framework_logcfg_file)

    # add kcwi specific config files # make changes here to allow this file
    # to be loaded from the command line
    if args.kcwi_config_file is None:
        kcwi_config_file = 'configs/kcwi.cfg'
        kcwi_config_fullpath = pkg_resources.resource_filename(
            pkg, kcwi_config_file)
        kcwi_config = ConfigClass(kcwi_config_fullpath, default_section='KCWI')
    else:
        # kcwi_config_fullpath = os.path.abspath(args.kcwi_config_file)
        kcwi_config = ConfigClass(args.kcwi_config_file, default_section='KCWI')

    # END HANDLING OF CONFIGURATION FILES ##########

    # Add current working directory to config info
    kcwi_config.cwd = os.getcwd()

    # check for the output directory
    check_directory(kcwi_config.output_directory)

    try:
        framework = Framework(Kcwi_pipeline, framework_config_fullpath)
        # add this line ONLY if you are using a local logging config file
        logging.config.fileConfig(framework_logcfg_fullpath)
        framework.config.instrument = kcwi_config
    except Exception as e:
        print("Failed to initialize framework, exiting ...", e)
        traceback.print_exc()
        sys.exit(1)
    framework.context.pipeline_logger = getLogger(framework_logcfg_fullpath,
                                                  name="KCWI")
    framework.logger = getLogger(framework_logcfg_fullpath, name="DRPF")

    if args.infiles is not None:
        framework.config.file_type = args.infiles

    # check for skipsky argument
    if args.skipsky:
        def_sk = getattr(framework.config.instrument, 'skipsky', None)
        if def_sk is not None:
            framework.context.pipeline_logger.info("Skipping sky subtraction")
            framework.config.instrument.skipsky = args.skipsky

    # check for taperfrac argument
    if args.taperfrac:
        def_tf = getattr(framework.config.instrument, 'TAPERFRAC', None)
        if def_tf is not None:
            framework.context.pipeline_logger.info(
                "Setting new taperfrac = %.3f" % args.taperfrac)
            framework.config.instrument.TAPERFRAC = args.taperfrac

    # check for middle_fraction argument
    if args.middle_fraction:
        def_mf = getattr(framework.config.instrument, 'MIDFRAC', None)
        if def_mf is not None:
            framework.context.pipeline_logger.info(
                "Setting new middle_fraction = %.2f" % args.middle_fraction)
            framework.config.instrument.MIDFRAC = args.middle_fraction

    # check for atlas_offset argument
    if args.atlas_offset:
        def_ao = getattr(framework.config.instrument, 'ATOFF', None)
        if def_ao is not None:
            framework.context.pipeline_logger.info(
                "Setting new atlas offset = %.2f" % args.atlas_offset)
            framework.config.instrument.ATOFF = args.atlas_offset

    # check for line_thresh argument
    if args.line_thresh:
        def_lt = getattr(framework.config.instrument, 'LINETHRESH', None)
        if def_lt is not None:
            framework.context.pipeline_logger.info(
                "Setting new line thresh = %.2f" % args.line_thresh)
            framework.config.instrument.LINETHRESH = args.line_thresh
    else:
        if args.blue:
            framework.config.instrument.LINETHRESH = float(
                kcwi_config.BLUE['linethresh'])
        elif args.red:
            framework.config.instrument.LINETHRESH = float(
                kcwi_config.RED['linethresh'])

    # check for tukey_alpha argument
    if args.tukey_alpha:
        def_ta = getattr(framework.config.instrument, 'TUKEYALPHA', None)
        if def_ta is not None:
            framework.context.pipeline_logger.info(
                "Setting new tukey alpha = %.2f" % args.tukey_alpha)
            framework.config.instrument.TUKEYALPHA = args.tukey_alpha
    else:
        if args.blue:
            framework.config.instrument.TUKEYALPHA = float(
                kcwi_config.BLUE['tukeyalpha'])
        elif args.red:
            framework.config.instrument.TUKEYALPHA = float(
                kcwi_config.RED['tukeyalpha'])

    # check for max_frac argument
    if args.max_frac:
        def_fm = getattr(framework.config.instrument, 'FRACMAX', None)
        if def_fm is not None:
            framework.context.pipeline_logger.info(
                "Setting new line windowing max fraction = %.2f" %
                args.max_frac)
            framework.config.instrument.FRACMAX = args.max_frac
    else:
        if args.blue:
            framework.config.instrument.FRACMAX = float(
                kcwi_config.BLUE['fracmax'])
        elif args.red:
            framework.config.instrument.FRACMAX = float(
                kcwi_config.RED['fracmax'])

    # check for atlas line list argument
    if args.atlas_line_list:
        def_ll = getattr(framework.config.instrument, 'LINELIST', None)
        if def_ll is not None:
            framework.context.pipeline_logger.info(
                "Using line list %s instead of generated list" %
                args.atlas_line_list)
            framework.config.instrument.LINELIST = args.atlas_line_list

    # update proc table argument
    if args.proctab:
        framework.context.pipeline_logger.info(
            "Using proc table file %s" % args.proctab
        )
        framework.config.instrument.procfile = args.proctab
    else:
        if args.blue:
            proctab = kcwi_config.BLUE['procfile']
        elif args.red:
            proctab = kcwi_config.RED['procfile']
        else:
            proctab = kcwi_config.procfile
        framework.context.pipeline_logger.info(
            "Using proc table file %s" % proctab)
        framework.config.instrument.procfile = proctab

    # set up channel specific parameters
    if args.blue:
        framework.config.instrument.arc_min_nframes = int(
            kcwi_config.BLUE['arc_min_nframes'])
        framework.config.instrument.contbars_min_nframes = int(
            kcwi_config.BLUE['contbars_min_nframes'])
        framework.config.instrument.object_min_nframes = int(
            kcwi_config.BLUE['object_min_nframes'])
        framework.config.instrument.minoscanpix = int(
            kcwi_config.BLUE['minoscanpix'])
        framework.config.instrument.oscanbuf = int(
            kcwi_config.BLUE['oscanbuf'])
    elif args.red:
        framework.config.instrument.arc_min_nframes = int(
            kcwi_config.RED['arc_min_nframes'])
        framework.config.instrument.contbars_min_nframes = int(
            kcwi_config.RED['contbars_min_nframes'])
        framework.config.instrument.object_min_nframes = int(
            kcwi_config.RED['object_min_nframes'])
        framework.config.instrument.minoscanpix = int(
            kcwi_config.RED['minoscanpix'])
        framework.config.instrument.oscanbuf = int(
            kcwi_config.RED['oscanbuf'])
    else:
        framework.config.instrument.arc_min_nframes = \
            kcwi_config.arc_min_nframes
        framework.config.instrument.contbars_min_nframes = \
            kcwi_config.contbars_min_nframes
        framework.config.instrument.object_min_nframes = \
            kcwi_config.object_min_nframes
        framework.config.instrument.minoscanpix = kcwi_config.minoscanpix
        framework.config.instrument.oscanbuf = kcwi_config.oscanbuf

    # start the bokeh server is requested by the configuration parameters
    if framework.config.instrument.enable_bokeh is True:
        if check_running_process(process='bokeh') is False:
            with open("bokeh_output.txt", "wb") as out:
                subprocess.Popen('bokeh serve', shell=True, stderr=out,
                                 stdout=out)
            # --session-ids=unsigned --session-token-expiration=86400',
            # shell=True)
            time.sleep(5)
        # subprocess.Popen('open http://localhost:5006?bokeh-session-id=kcwi',
        # shell=True)

    # initialize the proctab and read it
    framework.context.proctab = Proctab()
    framework.context.proctab.read_proctab(framework.config.instrument.procfile)

    framework.logger.info("Framework initialized")

    # add a start_bokeh event to the processing queue,
    # if requested by the configuration parameters
    if framework.config.instrument.enable_bokeh:
        framework.append_event('start_bokeh', None)

    # important: to be able to use the grouping mode, we need to reset
    # the default ingestion action to no-event,
    # otherwise the system will automatically trigger the next_file event
    # which initiates the processing
    if args.group_mode is True:
        # set the default ingestion event to None
        framework.config.default_ingestion_event = "add_only"

    # start queue manager only (useful for RPC)
    if args.queue_manager_only:
        # The queue manager runs forever.
        framework.logger.info("Starting queue manager only, no processing")
        framework.start(args.queue_manager_only)

    # in the next two ingest_data command, if we are using standard mode,
    # the first event generated is next_file.
    # if we are in groups mode (aka smart mode), then the first event generated
    # is no_event then, manually, a next_file event is generated for each group
    # specified in the variable imtypes

    # single frame processing
    elif args.frames:
        frames = []
        for frame in args.frames:
            # Verify we have the correct channel selected
            if args.blue and 'kr' in frame:
                print('Blue channel requested, but red files in list')
                qstr = input('Proceed? <cr>=yes or Q=quit: ')
                if 'Q' in qstr.upper():
                    frames = []
                    break
            if args.red and 'kb' in frame:
                print('Red channel requested, but blue files in list')
                qstr = input('Proceed? <cr>=yes or Q=quit: ')
                if 'Q' in qstr.upper():
                    frames = []
                    break
            frames.append(frame)
        framework.ingest_data(None, frames, False)

    # processing of a list of files contained in a file
    elif args.file_list:
        frames = []
        with open(args.file_list) as file_list:
            for frame in file_list:
                if "#" not in frame:
                    # Verify we have the correct channel selected
                    if args.blue and 'kr' in frame:
                        print('Blue channel requested, but red files in list')
                        qstr = input('Proceed? <cr>=yes or Q=quit: ')
                        if 'Q' in qstr.upper():
                            frames = []
                            break
                    if args.red and 'kb' in frame:
                        print('Red channel requested, but blue files in list')
                        qstr = input('Proceed? <cr>=yes or Q=quit: ')
                        if 'Q' in qstr.upper():
                            frames = []
                            break
                    frames.append(frame.strip('\n'))

        framework.ingest_data(None, frames, False)

        with open(args.file_list + '_ingest', 'w') as ingest_f:
            ingest_f.write('Files ingested at: ' +
                           datetime.datetime.now().isoformat())

    # ingest an entire directory, trigger "next_file" (which is an option
    # specified in the config file) on each file,
    # optionally continue to monitor if -m is specified
    elif args.dirname is not None:

        framework.ingest_data(args.dirname, None, args.monitor)

    # implement the group mode
    if args.group_mode is True:
        data_set = framework.context.data_set

        # remove focus images and ccd clearing images from the dataset
        focus_frames = data_set.data_table[data_set.data_table.OBJECT ==
                                           "focus"].index
        ccdclear_frames = data_set.data_table[data_set.data_table.OBJECT ==
                                              "Clearing ccd"].index
        data_set.data_table.drop(focus_frames, inplace=True)
        data_set.data_table.drop(ccdclear_frames, inplace=True)

        # processing
        imtypes = ['BIAS', 'CONTBARS', 'ARCLAMP', 'FLATLAMP', 'DOMEFLAT',
                   'TWIFLAT', 'OBJECT']

        for imtype in imtypes:
            subset = data_set.data_table[
                framework.context.data_set.data_table.IMTYPE == imtype]
            if 'OBJECT' in imtype:  # Ensure that standards are processed first
                object_order = []
                standard_order = []
                for frame in subset.index:
                    if is_file_kcwi_std(frame, logger=framework.context.logger):
                        standard_order.append(frame)
                    else:
                        object_order.append(frame)
                order = standard_order + object_order  # Standards first
                process_list(order)
            else:
                process_subset(subset)

    framework.config.instrument.wait_for_event = args.wait_for_event
    framework.config.instrument.continuous = args.continuous

    framework.start(args.queue_manager_only, args.ingest_data_only,
                    args.wait_for_event, args.continuous)


if __name__ == "__main__":
    main()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_writer

import numpy as np


class FlagSaturation(BasePrimitive):
    """
    Flag saturated pixels.

    Currently flags pixels with values > 60,000 counts with a value of 8
    in the flags FITS extension and updates the following FITS header keywords:

        * SATFLAG: set to ``True`` if operation is performed.
        * NSATFLAG: set to the count of saturated pixels.

    Uses the following configuration parameter:

        * saveintims: if set to ``True`` write out a flagged version of image in \*_fsat.fits.  Defaults to ``False``.

    Updates the flag extension of the image in the returned arguments.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _perform(self):
        self.logger.info("Flagging saturated pixels")

        # Header keyword to update
        key = 'SATFLAG'
        keycom = 'flagged saturated pixels?'

        # Create flags for saturated pixels
        flags = np.zeros(self.action.args.ccddata.data.shape, dtype=np.uint8)

        sat = np.where(self.action.args.ccddata.data > 60000)
        flags[sat] += 8
        number_of_sat_pixels = np.sum(sat)

        sat = np.where(self.action.args.ccddata.data == 0)
        flags[sat] += 8
        number_of_sat_pixels += np.sum(sat)

        self.action.args.ccddata.header[key] = (True, keycom)

        self.logger.info("Flagged %d saturated pixels" % number_of_sat_pixels)
        self.action.args.ccddata.header['NSATFLAG'] = \
            (number_of_sat_pixels, 'number of saturated pixels flagged')

        log_string = FlagSaturation.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.logger.info(log_string)

        # add flags array
        # DN 2023-may-28: commenting out mask update because
        # it causes bad things later on
        # self.action.args.ccddata.mask = flags
        self.action.args.ccddata.flags = flags

        if self.config.instrument.saveintims:
            kcwi_fits_writer(self.action.args.ccddata,
                             table=self.action.args.table,
                             output_file=self.action.args.name,
                             output_dir=self.config.instrument.output_directory,
                             suffix="fsat")

        return self.action.args
    # END: class CorrectDefects()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_reader, \
    kcwi_fits_writer, get_master_name, strip_fname

import os


class CorrectIllumination(BasePrimitive):
    """
    Correct for illumination and response variations using master flat.

    Currently, gives precedence for internal flats (cflat), as they are the
    most universally applicable.  To use other flats move the cflats aside, in
    which case any twilight flat (tflat) master would then have precedence
    followed by any dome flat (dflat) master.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _pre_condition(self):
        """
        Checks if we can correct illumination based on the processing table
        :return:
        """
        self.action.args.master_flat = None
        self.logger.info("Checking precondition for CorrectIllumination")
        # first check for internal flat
        target_type = 'MFLAT'
        tab = self.context.proctab.search_proctab(
            frame=self.action.args.ccddata, target_type=target_type,
            nearest=True)
        if len(tab) <= 0:
            # next look for twilight flat
            target_type = 'MTWIF'
            tab = self.context.proctab.search_proctab(
                frame=self.action.args.ccddata, target_type=target_type,
                nearest=True)
            if len(tab) <= 0:
                # finally look for dome flat
                target_type = 'MDOME'
                tab = self.context.proctab.search_proctab(
                    frame=self.action.args.ccddata, target_type=target_type,
                    nearest=True)
                if len(tab) <= 0:
                    precondition = False
                else:
                    precondition = True
            else:
                precondition = True
        else:
            precondition = True

        self.logger.info("pre condition got %d %s flats, expected 1"
                         % (len(tab), target_type))
        if precondition:
            self.action.args.master_flat = get_master_name(tab, target_type)
        return precondition

    def _perform(self):

        # Header keyword to update
        key = 'FLATCOR'
        keycom = 'flat corrected?'
        # obj, sky
        obj = None
        sky = None

        self.logger.info("Correcting Illumination")
        if self.action.args.master_flat:
            mflat = kcwi_fits_reader(
                os.path.join(self.config.instrument.cwd,
                             self.config.instrument.output_directory,
                             self.action.args.master_flat))[0]

            # do the correction
            self.action.args.ccddata.data *= mflat.data

            # update header keywords
            self.action.args.ccddata.header[key] = (True, keycom)
            self.action.args.ccddata.header['MFFILE'] = (
                self.action.args.master_flat, "Master flat filename")

            # check for obj, sky observations
            if self.action.args.nasmask and self.action.args.numopen > 1:
                ofn = self.action.args.name

                objfn = strip_fname(ofn) + '_obj.fits'
                full_path = os.path.join(
                    self.config.instrument.cwd,
                    self.config.instrument.output_directory, objfn)
                if os.path.exists(full_path):
                    obj = kcwi_fits_reader(full_path)[0]
                    # correction
                    obj.data *= mflat.data
                    # update header
                    obj.header[key] = (True, keycom)
                    obj.header['MFFILE'] = (
                        self.action.args.master_flat, 'Master flat filename')
                else:
                    obj = None

                skyfn = strip_fname(ofn) + '_sky.fits'
                full_path = os.path.join(
                    self.config.instrument.cwd,
                    self.config.instrument.output_directory, skyfn)
                if os.path.exists(full_path):
                    sky = kcwi_fits_reader(full_path)[0]
                    # correction
                    sky.data *= mflat.data
                    # update header
                    sky.header[key] = (True, keycom)
                    sky.header['MFFILE'] = (
                        self.action.args.master_flat, 'Master flat filename')
                else:
                    sky = None
        else:
            self.logger.error("No master flat found, "
                              "cannot correct illumination.")
            self.action.args.ccddata.header[key] = (False, keycom)

        log_string = CorrectIllumination.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string

        # write out intf image
        kcwi_fits_writer(self.action.args.ccddata,
                         table=self.action.args.table,
                         output_file=self.action.args.name,
                         output_dir=self.config.instrument.output_directory,
                         suffix="intf")
        self.context.proctab.update_proctab(frame=self.action.args.ccddata,
                                            suffix="intf",
                                            filename=self.action.args.name)
        self.context.proctab.write_proctab(tfil=self.config.instrument.procfile)

        # check for obj, sky images
        if obj is not None:
            kcwi_fits_writer(obj, output_file=self.action.args.name,
                             output_dir=self.config.instrument.output_directory,
                             suffix="objf")
        if sky is not None:
            kcwi_fits_writer(sky, output_file=self.action.args.name,
                             output_dir=self.config.instrument.output_directory,
                             suffix="skyf")

        self.logger.info(log_string)

        return self.action.args
    # END: class CorrectIllumination()



from keckdrpframework.models.arguments import Arguments
from astropy.io import fits
from astropy.nddata import CCDData
from astropy.table import Table
# from astropy import units as u
import numpy as np
from datetime import datetime

from keckdrpframework.primitives.base_primitive import BasePrimitive
import os
import logging
import pkg_resources
import subprocess
from pathlib import Path

logger = logging.getLogger('KCWI')

red_amp_dict = {'L1': 0, 'L2': 1, 'U1': 2, 'U2': 3}
# gains for slow readout, high gain
red_amp_gain = {'L1': 1.54, 'L2': 1.551, 'U1': 1.61, 'U2': 1.526}

# gains for slow readout gainmul by ampid
blue_amp_gain = {1:  {0: 1.570, 1: 1.600, 2: 1.610, 3: 1.600},
                 2:  {0: 0.785, 1: 0.800, 2: 0.805, 3: 0.800},
                 5:  {0: 0.314, 1: 0.320, 2: 0.325, 3: 0.319},
                 10: {0: 0.157, 1: 0.160, 2: 0.158, 3: 0.158}}


def parse_imsec(section=None):
    """
    Parse image section FITS header keyword into useful tuples.

    Take into account one-biased IRAF-style image section keywords and the
    possibility that a third element (strid) may be present and generate the
    x and y limits as well as the stride for each axis that are useful for
    python image slices.

    Args:
        section (str): square-bracket enclosed string with colon range
        specifiers and comma delimiters.

    :returns:
        - list: (int) y0, y1, x0, x1 - zero-biased (python) slice limits.
        - list: (int) y-stride, x-stride - strides for each axis.

    """
    xsec = section[1:-1].split(',')[0]
    ysec = section[1:-1].split(',')[1]
    xparts = xsec.split(':')
    yparts = ysec.split(':')
    p1 = int(xparts[0])
    p2 = int(xparts[1])
    p3 = int(yparts[0])
    p4 = int(yparts[1])
    # check for scale factor
    if len(xparts) == 3:
        xstride = int(xparts[2])
    else:
        xstride = 1
    if len(yparts) == 3:
        ystride = int(yparts[2])
    else:
        ystride = 1
    # is x axis in descending order?
    if p1 > p2:
        x0 = p2 - 1
        x1 = p1 - 1
        xstride = -abs(xstride)
    # x axis in ascending order
    else:
        x0 = p1 - 1
        x1 = p2 - 1
    # is y axis in descending order
    if p3 > p4:
        y0 = p4 - 1
        y1 = p3 - 1
        ystride = -abs(ystride)
    # y axis in ascending order
    else:
        y0 = p3 - 1
        y1 = p4 - 1
    # ensure no negative indices
    if x0 < 0:
        x0 = 0
    if y0 < 0:
        y0 = 0
    # package output with python axis ordering
    sec = (y0, y1, x0, x1)
    stride = (ystride, xstride)

    return sec, stride


def plotlabel(args):
    """Return label string for plot titles as (str)."""
    lab = "[ Img # %d " % args.ccddata.header['FRAMENO']
    lab += "(%s) " % args.illum
    lab += "Slicer: %s, " % args.ifuname
    lab += "Grating: %s, " % args.grating
    lab += "CWave: %d" % int(args.cwave)
    if 'BLUE' in args.ccddata.header['CAMERA'].upper():
        lab += ", Filter: %s " % args.filter
    lab += "] "
    return lab


class KCCDData(CCDData):
    """
    A container for KCWI images based on the CCDData object that adds
    the `noskysub` frame to allow parallel processing of sky-subtracted
    and un-sky-subtracted KCWI images.

    Attributes:
        noskysub (`numpy.ndarray` or None): Optional un-sky-subtracted frame
            that is created at the sky subtraction stage and processed in
            parallel with the primary frame.
            Default is ``None``.

    """

    def __init__(self, *args, **kwd):
        super().__init__(*args, **kwd)
        self._noskysub = None

    @property
    def noskysub(self):
        return self._noskysub

    @noskysub.setter
    def noskysub(self, value):
        self._noskysub = value


class ingest_file(BasePrimitive):
    """
    File ingestion class for KCWI images.

    Args:
        action (str): Pipeline action
        context: Pipeline context which includes arguments for the action.

    Attributes:
        logger: KCWI pipeline logger

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def get_keyword(self, keyword):
        """
        Get a keyword from ingested headers.

        Args:
            keyword (str): Keyword to fetch from header

        Returns:
            Keyword value if present, otherwise ``None``.
        """
        try:
            keyval = self.ccddata.header[keyword]
        except KeyError:
            keyval = None
        # return self.context.data_set.get_info_column(self.name, keyword)
        return keyval

    def camera(self):
        """
        Get camera ID number.

        Uses `CAMERA` FITS header keyword.

        Returns:
            0 for Blue channel, 1 for Red, and -1 for Unknown.
        """
        camera = self.get_keyword('CAMERA').upper()
        if 'BLUE' in camera:
            return 0
        elif 'RED' in camera:
            return 1
        else:
            return -1

    def camang(self):
        """
        Return the articulation stage camera angle.

        Determines which channel to use based on camera() call, then
        uses keyword `BARTANG` for Blue channel and `RARTANG` for Red.

        Raises:
            ValueError: if the camera ID is Unknown.

        Returns:
            (float): Camera angle in degrees for the given channel.

        """
        if self.camera() == 0:  # Blue
            key = 'BARTANG'
        elif self.camera() == 1:  # Red
            key = 'RARTANG'
        else:
            raise ValueError("unable to determine camera angle: "
                             "CAMERA undefined")
        return self.get_keyword(key)

    def filter(self):
        """
        Return the filter name.

        Determines which channel to use based on camera() call, then
        uses keyword `BFILTNAM` for Blue channel and returns `None` for Red.

        Raises:
            ValueError: if the camera ID is Unknown.

        Returns:
            (str): Filter name for the given channel.

        """
        if self.camera() == 0:  # Blue
            filt = self.get_keyword('BFILTNAM')
        elif self.camera() == 1:  # Red
            filt = 'None'
        else:
            raise ValueError("unable to determine filter: "
                             "CAMERA undefined")
        return filt

    def grangle(self):
        """
        Return the grating angle.

        Determines which channel to use based on camera() call, then
        uses keyword `BGRANGLE` for Blue channel and `RGRANGLE` for Red.

        Raises:
            ValueError: if the camera ID is Unknown.

        Returns:
            (float): Grating angle in degrees for the given channel.

        """
        if self.camera() == 0:  # Blue
            key = 'BGRANGLE'
        elif self.camera() == 1:  # Red
            key = 'RGRANGLE'
        else:
            raise ValueError("unable to determine grating angle: "
                             "CAMERA undefined")
        return self.get_keyword(key)

    def grating(self):
        """
        Return the grating name.

        Determines which channel to use based on camera() call, then
        uses keyword `BGRATNAM` for Blue channel and `RGRATNAM` for Red.

        Raises:
            ValueError: if the camera ID is Unknown.

        Returns:
            (str): Grating name for the given channel.

        """
        if self.camera() == 0:  # Blue
            key = 'BGRATNAM'
        elif self.camera() == 1:  # Red
            key = 'RGRATNAM'
        else:
            raise ValueError("unable to determine grating: CAMERA undefined")
        return self.get_keyword(key)

    def adjang(self):
        """
        Return the adjustment angle in degrees for the current grating.

        Returns:
            (float): 180.0 for high-resolution gratings, otherwise 0.0

        """
        if 'BH' in self.grating() or 'RH' in self.grating():
            return 180.0
        if 'BM' in self.grating() or 'RM' in self.grating():
            return 0.0
        if 'BL' in self.grating() or 'RL' in self.grating():
            return 0.0

    def atsig(self):
        """
        Return the Gaussian sigma to use for convolving the Atlas spectrum.

        Uses grating() and ifunum() to select a Gaussian sigma that when
        convolved with the Atlas spectrum will match the resolution of the
        observed spectrum.

        Raises:
            ValueError: if grating or ifunum canoot be determined.

        Returns:
            (float): sigma in pixels ranging from 0.75 to 14.0, or 0.0 if
            imtype() is `BIAS`.

        """
        if 'H' in self.grating():
            if self.ifunum() > 2:
                return 0.75
            else:
                return 2.5
        elif 'M' in self.grating():
            if self.ifunum() >= 2:
                return 4.0
            else:
                return 8.0
        elif 'L' in self.grating():
            if self.ifunum() == 2:
                return 10.0
            elif self.ifunum() == 3:
                return 5.0
            else:
                return 14.0
        else:
            if 'BIAS' in self.imtype():
                return 0.
            else:
                raise ValueError(
                    "unable to determine atlas sigma: GRATING undefined")

    def rho(self):
        """
        Return the `rho` value (lines/mm divided by 1000) for the given grating.

        Uses grating() to determine ingested grating.

        Returns:
            (float, or ``None``): `rho` value or ``None`` if grating unknown.

        """
        if 'BH1' in self.grating():
            return 3.6
        elif 'BH2' in self.grating():
            return 3.2805
        elif 'BH3' in self.grating():
            return 2.8017
        elif 'RH1' in self.grating():
            return 2.360
        elif 'RH2' in self.grating():
            return 2.020
        elif 'RH3' in self.grating():
            return 1.735
        elif 'RH4' in self.grating():
            return 1.490
        elif 'BM' in self.grating():
            return 1.901
        elif 'RM1' in self.grating():
            return 1.300
        elif 'RM2' in self.grating():
            return 0.945
        elif 'BL' in self.grating():
            return 0.870
        elif 'RL' in self.grating():
            return 0.530
        else:
            return None

    def cwave(self):
        """
        Return the central wavelength in Angstroms.

        Determines which channel to use based on camera() call, then
        uses keyword `BCWAVE` for Blue channel and `RCWAVE` for Red.

        Raises:
            ValueError: if the camera ID is Unknown.

        Returns:
            (float): Central wavelength in Angstroms for the given channel.

        """
        if self.camera() == 0:  # Blue
            key = 'BCWAVE'
        elif self.camera() == 1:  # Red
            key = 'RCWAVE'
        else:
            raise ValueError("unable to determine central wavelength: "
                             "CAMERA undefined")
        return self.get_keyword(key)

    def dich(self):
        """
        Query if dichroic present for image.

        Returns ``True`` for all Red channel data, but checks for the presence
        of Red keywords in the Blue channel header to determine status of
        dichroic.

        Returns:
            (bool): ``True`` if dichroic present, ``False`` if not.

        """
        if self.camera() == 0:  # Blue
            if self.get_keyword('RCWAVE'):
                return True
            else:
                return False
        elif self.camera() == 1:    # Red
            return True

    def resolution(self):
        """
        Return FWHM resolution in Angstroms for the given grating.

        Calls cwave(), grating(), and ifunum() to calculate resolution.

        Raises:
            ValueError: if the grating is unknown.

        Returns:
            (float): FWHM in Angstroms, or 0. if imtype() is `BIAS`.

        """
        # get reference wavelength
        refwave = self.cwave()
        if refwave:
            rw = refwave
        else:
            if 'B' in self.grating():
                rw = 4500.
            else:
                rw = 7500.
        # Calc rez from grating resolution (d-lambda = lambda/R)
        # First, assume large slicer IFU
        if 'BH' in self.grating():
            rez = rw / 5000.
        elif 'RH' in self.grating():
            rez = rw / 4600.
        elif 'BM' in self.grating():
            rez = rw / 2500.
        elif 'RM' in self.grating():
            rez = rw / 1900.
        elif 'BL' in self.grating():
            rez = rw / 1250.
        elif 'RL' in self.grating():
            rez = rw / 800.
        else:
            # Resolution not needed for biases
            if 'BIAS' in self.imtype():
                rez = 0.
            else:
                raise ValueError("unable to compute atlas resolution: "
                                 "grating undefined")
        # Adjust for slicer
        if self.ifunum() == 2:  # Medium slicer
            rez /= 2.
        elif self.ifunum() == 3:  # Small slicer
            rez /= 4.

        return rez

    def delta_wave_out(self):
        """
        Return output delta lambda in Angstroms for the given grating

        Calls grating(), and ybinsize() to calculate output delta lambda.

        Raises:
            ValueError: if the grating is unknown.

        Returns:
            (float): delta lambda in Angstroms, or 0. if imtype() is `BIAS`.

        """
        # Calc delta wave out from grating
        if 'BH' in self.grating():
            dw = 0.125 * float(self.ybinsize())
        elif 'RH' in self.grating():
            dw = 0.125 * float(self.ybinsize())
        elif 'BM' in self.grating():
            dw = 0.25 * float(self.ybinsize())
        elif 'RM' in self.grating():
            dw = 0.25 * float(self.ybinsize())
        elif 'BL' in self.grating():
            dw = 0.5 * float(self.ybinsize())
        elif 'RL' in self.grating():
            dw = 0.5 * float(self.ybinsize())
        else:
            # Delta wave not needed for biases
            if 'BIAS' in self.imtype():
                dw = 0.
            else:
                raise ValueError("unable to compute output delta lambda: "
                                 "grating undefined")
        return dw

    def namps(self):
        """
        Return value of `NVIDINP` FITS header keyword.
        """
        return self.get_keyword('NVIDINP')

    def nasmask(self):
        """
        Query if mask was inserted for image.

        Calls camera() to determine channel, and then uses `BNASNAM` for Blue
        channel or `RNASNAM` for Red channel.

        Raises:
            ValueError: if channel is Unknown.

        Returns:
            (bool): ``True`` if 'Mask' in, ``False`` if not.

        """
        if self.camera() == 0:  # Blue
            if 'Mask' in self.get_keyword('BNASNAM'):
                return True
            else:
                return False
        elif self.camera() == 1:  # Red
            if 'Mask' in self.get_keyword('RNASNAM'):
                return True
            else:
                return False
        else:
            raise ValueError("unable to determine mask: CAMERA undefined")

    def numopen(self):
        """Returns value of `NUMOPEN` FITS header keyword."""
        return self.get_keyword('NUMOPEN')

    def shufrows(self):
        """Returns value of `SHUFROWS` FITS header keyword."""
        return self.get_keyword('SHUFROWS')

    def ampmode(self):
        """Returns value of `AMPMODE` FITS header keyword."""
        return self.get_keyword('AMPMODE')

    def xbinsize(self):
        """Return X part of `BINNING` keyword value as (int)."""
        return int(self.get_keyword('BINNING').split(',')[0])

    def ybinsize(self):
        """Return Y part of `BINNING` keyword value as (int)."""
        return int(self.get_keyword('BINNING').split(',')[-1])

    def plotlabel(self):
        """Return label string for plot titles as (str)."""
        lab = "[ Img # %d " % self.get_keyword('FRAMENO')
        lab += "(%s) " % self.illum()
        lab += "Slicer: %s, " % self.ifuname()
        lab += "Grating: %s, " % self.grating()
        lab += "CWave: %d, " % int(self.cwave())
        lab += "Filter: %s " % self.filter()
        lab += "] "
        return lab

    def stdlabel(self):
        """Return label string for standard star plot titles as (str)."""
        lab = "[ Img # %d " % self.get_keyword('FRAMENO')
        lab += "(%s) " % self.get_keyword('TARGNAME')
        lab += "Slicer: %s, " % self.ifuname()
        lab += "Grating: %s, " % self.grating()
        lab += "CWave: %d, " % int(self.cwave())
        lab += "Filter: %s " % self.filter()
        lab += "] "
        return lab

    def ifuname(self):
        """Return the value of the `IFUNAM` FITS header keyword."""
        return self.get_keyword('IFUNAM')

    def ifunum(self):
        """Return the value of the `IFUNUM` FITS header keyword."""
        return self.get_keyword('IFUNUM')

    def imtype(self):
        """Return the value of the `IMTYPE` FITS header keyword."""
        return self.get_keyword('IMTYPE')

    def illum(self):
        """
        Generate a string that characterizes the illumination for the frame.

        Uses various FITS header keywords to determine the kind of illumination
        that was used for the frame.  If a consistent picture of the
        illumination cannot be determined, set the return value to `Test`.

        Returns:
            (str): Characterization of illumination for given frame.

        """
        # ARCS
        if self.get_keyword('IMTYPE') == 'ARCLAMP':
            if self.get_keyword('LMP0STAT') == 1 and \
                    self.get_keyword('LMP0SHST') == 1:
                illum = self.get_keyword('LMP0NAM')
            elif self.get_keyword('LMP1STAT') == 1 and \
                    self.get_keyword('LMP1SHST') == 1:
                illum = self.get_keyword('LMP1NAM')
            else:
                illum = 'Test'
        # Internal FLATS
        elif self.get_keyword('IMTYPE') == 'FLATLAMP':
            if self.get_keyword('LMP3STAT') == 1:
                illum = 'Contin'
            else:
                illum = 'Test'
        # DOMES
        elif self.get_keyword('IMTYPE') == 'DOMEFLAT':
            if self.get_keyword('FLIMAGIN') == 'on' or \
                    self.get_keyword('FLSPECTR') == 'on':
                illum = 'Dome'
            else:
                illum = 'Test'
        # Twilight FLATS
        elif self.get_keyword('IMTYPE') == 'TWIFLAT':
            illum = 'Twilit'
        # BARS
        elif self.get_keyword('IMTYPE') == 'CONTBARS':
            if self.get_keyword('LMP3STAT') == 1:
                illum = 'Contin'
            else:
                illum = 'Test'
        # OBJECT
        elif self.get_keyword('IMTYPE') == 'OBJECT':
            obnam = self.get_keyword('OBJECT')
            if obnam is None:
                obnam = self.get_keyword('TARGNAME')
                if obnam is None:
                    obnam = 'Object'
            # clean up the string
            illum = obnam.replace("/", "_").replace(" ", "").replace(".", "_")
        else:
            illum = 'Test'
        return illum

    def calibration_lamp(self):
        """
        Determine which calibration source was used for a given frame.

        Examines `LMPnSTAT` and `LMPnSHST` keywords to determine which lamp was
        on and which shutter was open and thus providing illuminate for the
        frame.  Returns ``None`` if not an `ARCLAMP` image type.

        Returns:
            (str, or ``None``): Which calibration lamp was active for frame.

        """
        if self.get_keyword('IMTYPE') != 'ARCLAMP':
            return None
        else:
            lamps_dictionary = {
                0: "FeAr",
                1: "ThAr",
                2: "Aux",
                3: "Continuum A"
            }
            for key in lamps_dictionary.keys():
                status = self.get_keyword('LMP%dSTAT' % key)
                shutter = self.get_keyword('LMP%dSHST' % key)
                if status == 1 and shutter == 1:
                    return lamps_dictionary[key]

    def map_ccd(self, xbin, ybin):
        """
        Return CCD section variables useful for processing

        Args:
            xbin (int): binning in x
            ybin (int): binning in y

        Uses FITS keyword NVIDINP to determine how many amplifiers were used
        to read out the CCD.  Then reads the corresponding BSECn, and
        DSECn keywords, where n is the amplifier number.  The indices are
        converted to Python (0-biased, y axis first) indices and an array
        is constructed for each of the two useful sections of the CCD as
        follows:

        * Bsec[0][0] - First amp, y lower limit
        * Bsec[0][1] - First amp, y upper limit
        * Bsec[0][2] - First amp, x lower limit
        * Bsec[0][3] - First amp, x upper limit
        * Bsec[1][0] - Second amp, y lower limit
        * etc.

        Bsec is the full overscan region for the given amplifier and is used
        to calculate and perform the overscan subtraction.

        Dsec is the full CCD region for the given amplifier and is used to
        trim the image after overscan subtraction has been performed.

        Tsec accounts for trimming the image according to Dsec.

        Amps are assumed to be organized as follows:

        .. code-block:: text

                      BLUE                          RED
            (0,ny)  --------- (nx,ny)    (0,ny)  --------- (nx,ny)
                    | 2 | 3 |                    | 0 | 2 |
                    ---------                    ---------
                    | 0 | 1 |                    | 1 | 3 |
            (0,0)   --------- (nx, 0)    (0,0)   --------- (nx, 0)

        :returns:
            - list: (int) y0, y1, x0, x1 for bias section
            - list: (int) y0, y1, x0, x1 for data section
            - list: (int) y0, y1, x0, x1 for trimmed section
            - list: (bool) y-direction, x-direction, True if forward, else False

        """

        namps = self.namps()    # int(self.get_keyword('NVIDINP'))
        # TODO: check namps
        camera = self.get_keyword('CAMERA').upper()
        ampmode = self.get_keyword('AMPMODE')
        # section lists
        bsec = []
        dsec = []
        tsec = []
        strides = []
        amps = []
        if 'BLUE' in camera:
            nb = 1    # numbering bias (0 or 1)
            # loop over amps
            for i in range(namps):
                ia = i + 1
                amps.append(ia)
                section = self.get_keyword('BSEC%d' % ia)
                sec, stride = parse_imsec(section)
                bsec.append(sec)
                section = self.get_keyword('DSEC%d' % ia)
                sec, stride = parse_imsec(section)
                dsec.append(sec)
                section = self.get_keyword('CSEC%d' % ia)
                sec, stride = parse_imsec(section)
                strides.append(stride)
                if i == 0:
                    y0 = 0
                    y1 = int((sec[1] - sec[0]) / ybin)
                    x0 = 0
                    x1 = int((sec[3] - sec[2]) / xbin)
                elif i == 1:
                    y0 = 0
                    y1 = int((sec[1] - sec[0]) / ybin)
                    x0 = tsec[0][3] + 1
                    x1 = x0 + int((sec[3] - sec[2]) / xbin)
                elif i == 2:
                    y0 = tsec[0][1] + 1
                    y1 = y0 + int((sec[1] - sec[0]) / ybin)
                    x0 = 0
                    x1 = int((sec[3] - sec[2]) / xbin)
                elif i == 3:
                    y0 = tsec[0][1] + 1
                    y1 = y0 + int((sec[1] - sec[0]) / ybin)
                    x0 = tsec[0][3] + 1
                    x1 = x0 + int((sec[3] - sec[2]) / xbin)
                else:
                    # should not get here
                    y0 = -1
                    y1 = -1
                    x0 = -1
                    x1 = -1
                    # self.log.info("ERROR - bad amp number: %d" % i)
                tsec.append((y0, y1, x0, x1))
        elif 'RED' in camera:
            nb = 0    # numbering bias (0 or 1)
            amp_count = 0
            for amp in red_amp_dict.keys():
                if amp in ampmode:
                    amp_count += 1
                    i = red_amp_dict[amp]
                    amps.append(i)
                    section = self.get_keyword('BSEC%d' % i)
                    sec, stride = parse_imsec(section)
                    bsec.append(sec)
                    section = self.get_keyword('DSEC%d' % i)
                    sec, stride = parse_imsec(section)
                    dsec.append(sec)
                    section = self.get_keyword('CSEC%d' % i)
                    sec, stride = parse_imsec(section)
                    strides.append(stride)
                    y0, y1, x0, x1 = sec
                    y0 = int(y0 / abs(stride[0]))
                    y1 = int(y1 / abs(stride[0]))
                    x0 = int(x0 / abs(stride[1]))
                    x1 = int(x1 / abs(stride[1])) - (abs(stride[1]) - 1)
                    tsec.append((y0, y1, x0, x1))
                else:
                    bsec.append((0, 0, 0, 0))
                    dsec.append((0, 0, 0, 0))
                    strides.append((1, 1))
                    tsec.append((0, 0, 0, 0))
            if amp_count != namps:
                self.logger.warning("Didn't get all the amps: %d", amp_count)
        else:
            self.logger.warning("Unknown CAMERA: %s" % camera)
            nb = 0

        return bsec, dsec, tsec, strides, amps, nb

    def _perform(self):
        # if self.context.data_set is None:
        #    self.context.data_set = DataSet(None, self.logger, self.config,
        #    self.context.event_queue)
        # self.context.data_set.append_item(self.action.args.name)
        self.logger.info(
            "------------------- Ingesting file %s -------------------" %
            self.action.args.name)

        self.action.event._recurrent = True

        self.name = self.action.args.name
        out_args = Arguments()

        ccddata, table = kcwi_fits_reader(self.name)

        # Are we already in proctab?
        out_args.in_proctab = self.context.proctab.in_proctab(frame=ccddata)
        if out_args.in_proctab:
            out_args.last_suffix = self.context.proctab.last_suffix(
                frame=ccddata)
            if len(out_args.last_suffix) > 0:
                self.logger.info("Last suffix is %s" % out_args.last_suffix)
                out_dir = self.config.instrument.output_directory
                last_file = os.path.join(out_dir,
                                         self.name.split('.fits')[0] + '_' +
                                         out_args.last_suffix + '.fits')
                if 'icubes' not in out_args.last_suffix:
                    self.logger.info("Ingest stub for %s" % last_file)
                    # ccddata, table = kcwi_fits_reader(last_file)
                else:
                    self.logger.info("Processing completed in %s" % last_file)
        else:
            out_args.last_suffix = ""

        # save the ccd data into an object
        # that can be shared across the functions
        self.ccddata = ccddata

        out_args.ccddata = ccddata
        out_args.table = table

        imtype = self.get_keyword("IMTYPE")
        groupid = self.get_keyword("GROUPID")

        if groupid is None:
            groupid = "NONE"
        else:
            if len(groupid) <= 0:
                groupid = "NONE"
        if imtype is None:
            fname = os.path.basename(self.action.args.name)
            self.logger.warn(f"Unknown IMTYPE {fname}")
            self.action.event._reccurent = False

        if self.check_if_file_can_be_processed(imtype) is False:

            if self.config.instrument.continuous or \
                    self.config.instrument.wait_for_event:
                self.logger.warn("Input frame cannot be reduced. Rescheduling")
                self.action.new_event = None
                return None
            else:
                self.logger.warn("Input frame cannot be reduced. Exiting")
                self.action.new_event = None
                self.action.event._recurrent = False
                return None
        else:
            self.action.event._recurrent = False

        out_args.name = self.action.args.name
        out_args.imtype = imtype
        out_args.groupid = groupid
        # CAMERA
        out_args.camera = self.camera()
        # DICH
        out_args.dich = self.dich()
        # CAMANGLE
        out_args.camangle = self.camang()
        # FILTER
        out_args.filter = self.filter()
        # GRANGLE
        out_args.grangle = self.grangle()
        # GRATING
        out_args.grating = self.grating()
        # ADJANGLE
        out_args.adjang = self.adjang()
        # RHO
        out_args.rho = self.rho()
        # CWAVE
        out_args.cwave = self.cwave()
        # RESOLUTION
        out_args.resolution = self.resolution()
        # ATSIG
        out_args.atsig = self.atsig()
        # DELTA WAVE OUT
        out_args.dwout = self.delta_wave_out()
        # NAMPS
        out_args.namps = self.namps()
        # NASMASK
        out_args.nasmask = self.nasmask()
        # SHUFROWS
        out_args.shufrows = self.shufrows()
        # NUMOPEN
        out_args.numopen = self.numopen()
        # AMPMODE
        out_args.ampmode = self.ampmode()
        # BINNING
        out_args.xbinsize, out_args.ybinsize = \
            map(int, self.get_keyword('BINNING').split(','))
        # IFUNUM
        out_args.ifunum = int(self.get_keyword('IFUNUM'))
        # IFUNAM
        out_args.ifuname = self.get_keyword('IFUNAM')
        # PLOTLABEL
        out_args.plotlabel = self.plotlabel()
        # STDLABEL
        out_args.stdlabel = self.stdlabel()
        # ILUM
        out_args.illum = self.illum()
        # MAPCCD
        out_args.map_ccd = self.map_ccd(out_args.xbinsize, out_args.ybinsize)
        # CALIBRATION LAMP
        out_args.calibration_lamp = self.calibration_lamp()
        # TTIME
        out_args.ttime = self.get_keyword('TTIME')

        return out_args

    def apply(self):
        """
        Apply method for class.

        Checks _pre_condition().  If ``True``, then call _perform() and
        collect output.  Then check _post_condition().

        Returns:
            output from _perform, or ``None`` if there was an exception.

        """
        if self._pre_condition():
            try:
                output = self._perform()
            except ValueError as e:
                self.logger.warn("UNABLE TO INGEST THE FILE")
                self.logger.warn("Reason: %s" % e)
                return None
            if self._post_condition():
                self.output = output
        return self.output

    def check_if_file_can_be_processed(self, imtype):
        """
        For a given image type, ensure that processing can proceed.

        Based on `IMTYPE` keyword, makes a call to proctab to see if
        pre-requisite images are present.

        Returns:
            (bool): ``True`` if processing can proceed, ``False`` if not.

        """
        if imtype == 'OBJECT':
            # bias frames
            bias_frames = self.context.proctab.search_proctab(
                frame=self.ccddata, target_type='MBIAS', nearest=True)
            # arclamp
            arclamp_frames = self.context.proctab.search_proctab(
                frame=self.ccddata, target_type='MARC', nearest=True)
            # master flats
            masterflat_frames = self.context.proctab.search_proctab(
                frame=self.ccddata, target_type='MFLAT', nearest=True)
            masterdome_frames = self.context.proctab.search_proctab(
                frame=self.ccddata, target_type='MDOME', nearest=True)
            mastertwif_frames = self.context.proctab.search_proctab(
                frame=self.ccddata, target_type='MTWIF', nearest=True)

            if len(bias_frames) > 0 and len(arclamp_frames) > 0 and (
                    len(masterflat_frames) > 0 or len(masterdome_frames) > 0 or
                    len(mastertwif_frames) > 0):
                return True
            else:
                self.logger.warn("Cannot reduce OBJECT frame. Found:")
                self.logger.warn(f"\tMBIAS: {len(bias_frames)}")
                self.logger.warn(f"\tMARC: {len(arclamp_frames)}")
                self.logger.warn(f"\tMFLAT: {len(masterflat_frames)}")
                self.logger.warn(f"\tMDOME: {len(masterdome_frames)}")
                self.logger.warn(f"\tMTWIF: {len(mastertwif_frames)}")

                return False

        if imtype == 'ARCLAMP':
            # continuum bars
            contbars_frames = self.context.proctab.search_proctab(
                frame=self.ccddata, target_type='MCBARS', nearest=True)
            if len(contbars_frames) > 0:
                return True
            else:
                self.logger.warn("Cannot reduce ARCLAMP frame. "
                                 "Missing master continuum bars. ")
                return False

        if imtype in ['FLATLAMP', 'TWIFLAT', 'DOMEFLAT']:
            # bias frames
            bias_frames = self.context.proctab.search_proctab(
                frame=self.ccddata, target_type='MBIAS', nearest=True)
            arc_frames = self.context.proctab.search_proctab(
                frame=self.ccddata, target_type='MARC', nearest=True)
            if len(bias_frames) > 0 and len(arc_frames) > 0:
                return True
            else:
                self.logger.warn(f"Cannot reduce {imtype} frame.")
                if len(bias_frames) <= 0:
                    self.logger.warn(f"Missing master bias.")
                if len(arc_frames) <= 0:
                    self.logger.warn(f"Missing master arc")
                return False

        return True


def kcwi_fits_reader(file):
    """A reader for KCCDData objects.

    Currently, this is a separate function, but should probably be
    registered as a reader similar to fits_ccddata_reader.

    Args:
        file (str): The filename (or pathlib.Path) of the FITS file to open.

    Raises:
        FileNotFoundError: if file not found or
        OSError: if file not accessible

    Returns:
        (KCCDData, FITS table): All relevant frames in a single KCCDData object
        and a FITS table of exposure events, if present otherwise ``None``.

    """
    try:
        hdul = fits.open(file)
    except (FileNotFoundError, OSError) as e:
        print(e)
        raise e
    read_imgs = 0
    read_tabs = 0
    # primary image
    ccddata = KCCDData(hdul['PRIMARY'].data, meta=hdul['PRIMARY'].header,
                       unit='adu')
    read_imgs += 1
    # check for other legal components
    if 'UNCERT' in hdul:
        ccddata.uncertainty = hdul['UNCERT'].data
        read_imgs += 1
    if 'FLAGS' in hdul:
        ccddata.flags = hdul['FLAGS'].data
        read_imgs += 1
    if 'MASK' in hdul:
        ccddata.mask = hdul['MASK'].data
        read_imgs += 1
    if 'NOSKYSUB' in hdul:
        ccddata.noskysub = hdul['NOSKYSUB'].data
        read_imgs += 1
    if 'Exposure Events' in hdul:
        table = hdul['Exposure Events']
        read_tabs += 1
    else:
        table = None
    # prepare for floating point
    ccddata.data = ccddata.data.astype(np.float64)
    # Fix red headers
    fix_header(ccddata)
    # Check for CCDCFG keyword
    if 'CCDCFG' not in ccddata.header:
        ccdcfg = ccddata.header['CCDSUM'].replace(" ", "")
        ccdcfg += "%1d" % ccddata.header['CCDMODE']
        ccdcfg += "%02d" % ccddata.header['GAINMUL']
        ccdcfg += "%02d" % ccddata.header['AMPMNUM']
        ccddata.header['CCDCFG'] = ccdcfg

    if ccddata:
        if 'BUNIT' in ccddata.header:
            ccddata.unit = ccddata.header['BUNIT']
            if ccddata.uncertainty:
                ccddata.uncertainty.unit = ccddata.header['BUNIT']
            # print("setting image units to " + ccddata.header['BUNIT'])

    logger.info("<<< read %d imgs and %d tables out of %d hdus in %s" %
                (read_imgs, read_tabs, len(hdul), file))
    return ccddata, table


def write_table(output_dir=None, table=None, names=None, comment=None,
                keywords=None, output_name=None, clobber=False):
    """
    Write out FITS table.

    Args:
        output_dir (str): output directory for table.
        table (list of arrays): each array in list should have the same size.
        names (list of str): one for each column in `table`.
        comment (string): text for FITS COMMENT header record.
        keywords (FITS keyword dict): for FITS header records.
        output_name (str): name for output table.
        clobber (bool): set to to ``True`` to overwrite existing table.

    Raises:
        FileExistsError: if cannot overwrite file or
        OSError: if some other access exception occurred.

    """
    output_file = os.path.join(output_dir, output_name)
    # check if file already exists
    if os.path.exists(output_file) and clobber is False:
        logger.warning("Table %s already exists and overwrite is set to False"
                       % output_file)
        return
    if os.path.exists(output_file) and clobber is True:
        logger.warning("Table %s already exists and will be overwritten"
                       % output_file)
        logger.info("Removing file: %s" % output_file)
        os.remove(output_file)

    t = Table(table, names=names)
    if comment:
        t.meta['COMMENT'] = comment
    if keywords:
        for k, v in keywords.items():
            t.meta[k] = v
    try:
        t.write(output_file, format='fits')
        logger.info("Output table: %s" % output_file)
    except (FileExistsError, OSError):
        logger.error("Something went wrong creating the table: "
                     "table already exists")


def read_table(input_dir=None, file_name=None):
    """
    Read FITS table

    Uses astropy.Table module to read in FITS table.

    Raises:
        FileNotFoundError: if file not found.

    Returns:
        (FITS Table): table read in or ``None`` if unsuccessful.

    """
    # Set up return table
    input_file = os.path.join(input_dir, file_name)
    logger.info("Trying to read table: %s" % input_file)
    try:
        retab = Table.read(input_file, format='fits')
    except FileNotFoundError:
        logger.warning("No table to read")
        retab = None
    return retab


def kcwi_fits_writer(ccddata, table=None, output_file=None, output_dir=None,
                     suffix=None):
    """
    A writer for KCCDData or CCDData objects.

    Updates history in FITS header with pipeline version and git repo version
    and date.

    Converts float64 data to float32.

    Uses object to_hdu() method to generate hdu list and then checks if various
    extra frames are present (flags, noskysub) and adds them to the hdu list
    prior to writing out with hdu list writeto() method.

    Note:
        Currently fits tables are not written out.

    Args:
        ccddata (KCCDData or CCDData): object to write out.
        table (FITS Table): currently not used.
        output_file (str): base filename to write to.
        output_dir (str): directory into which to write.
        suffix (str): a suffix to append to output_file string.


    """
    # Determine if the version info is already in the header
    contains_version = False
    for h in ccddata.header["HISTORY"]:
        if "kcwidrp version" in h:
            contains_version = True

    if not contains_version:
        # Add setup.py version number to header
        version = pkg_resources.get_distribution('kcwidrp').version
        ccddata.header.add_history(f"kcwidrp version={version}")

        # Get string filepath to .git dir, relative to this primitive
        primitive_loc = os.path.dirname(os.path.abspath(__file__))
        git_loc = primitive_loc[:-18] + ".git"

        # Attempt to gather git version information
        git1 = subprocess.run(["git", "--git-dir", git_loc, "describe",
                               "--tags", "--long"], capture_output=True)
        git2 = subprocess.run(["git", "--git-dir", git_loc, "log", "-1",
                               "--format=%cd"], capture_output=True)
        
        # If all went well, save to the header
        if not bool(git1.stderr) and not bool(git2.stderr):
            git_v = git1.stdout.decode('utf-8')[:-1]
            git_d = git2.stdout.decode('utf-8')[:-1]
            ccddata.header.add_history(f"git version={git_v}")
            ccddata.header.add_history(f"git date={git_d}")
        else:
            logger.debug("Package not installed from a git repo, skipping")

    # If there is a data array, and the type of that array is a 64-bit float,
    # force it to 32 bits.
    if ccddata.data is not None and ccddata.data.dtype == np.float64:
        ccddata.data = ccddata.data.astype(np.float32)
    # If there is an uncertainty array, and the values within
    # (the .array property), make it 32 bits.
    if ccddata.uncertainty is not None and ccddata.uncertainty.array.dtype == np.float64:
        ccddata.uncertainty.array = ccddata.uncertainty.array.astype(np.float32)
    
    out_file = os.path.join(output_dir, os.path.basename(output_file))
    if suffix is not None:
        (main_name, extension) = os.path.splitext(out_file)
        out_file = main_name + "_" + suffix + extension
    hdus_to_save = ccddata.to_hdu()
    # check for flags
    flags = getattr(ccddata, "flags", None)
    if flags is not None:
        hdus_to_save.append(fits.ImageHDU(flags, name='FLAGS',
                                          do_not_scale_image_data=True))
    # check for noskysub
    nskysb = getattr(ccddata, "noskysub", None)
    if nskysb is not None:
        if ccddata.noskysub.dtype == np.float64:
            ccddata.noskysub = ccddata.noskysub.astype(np.float32)
        hdus_to_save.append(fits.ImageHDU(nskysb, name='NOSKYSUB'))
    # something about the way the original table is written out is wrong
    # and causes problems.  Leaving it off for now.
    # if table is not None:
    #    hdus_to_save.append(table)
    # log
    logger.info(">>> Saving %d hdus to %s" % (len(hdus_to_save), out_file))
    hdus_to_save.writeto(out_file, overwrite=True)


def strip_fname(filename):
    """
    Return pathlib.Path.stem attribute for given filename.

    Args:
        filename (str): filename to strip.

    Returns:
        (str): Path(filename).stem, or filename if error occurs.

    """
    if not filename:
        logger.error(f"Failed to strip file {filename}")
        return filename
    strip = Path(filename).stem
    return strip


def get_master_name(tab, target_type, loc=0):
    """
    Add a specific tag to an output fits filename read from a proc table.

    Args:
        tab (proc table): proc table source of filename.
        target_type (str): suffix to add after underscore.
        loc (int): row within table in `tab`, defaults to 0.

    Returns:
        (str): constructed filename from input tab entry and target_type.

    """
    res = Path(strip_fname(tab['filename'][loc]) + '_' +
               target_type.lower() + ".fits").name
    return res


def master_bias_name(ccddata, target_type='MBIAS'):
    # Currently NOT USED (DN, 7-Sep-2023)
    # Delivers a mbias filename that is unique for each CCD configuration
    # Any KCWI frame with a shared CCD configuration can use the same bias
    name = target_type.lower() + '_' + ccddata.header['CCDCFG'] + '.fits'
    return name


def master_flat_name(ccddata, target_type):
    # Currently NOT USED (DN, 7-Sep-2023)
    # Delivers a name that is unique across an observing block
    name = target_type.lower() + '_' + ccddata.header['STATEID'] + '.fits'
    return name


def fix_header(ccddata):
    """
    Fix header keywords for DRP use.

    Update GAINn keywords for Blue channel.

    Add FITS header keywords to Red channel data to make compatible with DRP.

    Adds the following keywords that are not present in raw images:

    * MJD - Modified Julian Day (only for AIT data)
    * NVIDINP - from TAPLINES keyword
    * GAINMUL - set to 1
    * CCDMODE - from CDSSPEED keyword
    * AMPNUM - based on AMPMODE
    * GAINn - from red_amp_gain dictionary

    """
    # are we blue?
    if 'BLUE' in ccddata.header['CAMERA'].upper():
        gainmul = ccddata.header['GAINMUL']
        namps = ccddata.header['NVIDINP']
        for ia in range(namps):
            ampid = ccddata.header['AMPID%d' % (ia+1)]
            gain = blue_amp_gain[gainmul][ampid]
            ccddata.header['GAIN%d' % (ia+1)] = gain
    # are we red?
    elif 'RED' in ccddata.header['CAMERA'].upper():
        # Fix red headers during Caltech AIT
        if 'TELESCOP' not in ccddata.header:
            # Add DCS keywords
            ccddata.header['TARGNAME'] = ccddata.header['OBJECT']
            dateend = ccddata.header['DATE-END']
            de = datetime.fromisoformat(dateend)
            day_frac = de.hour / 24. + de.minute / 1440. + de.second / 86400.
            jd = datetime.date(de).toordinal() + day_frac + 1721424.5
            mjd = jd - 2400000.5
            ccddata.header['MJD'] = mjd
        # Add NVIDINP
        ccddata.header['NVIDINP'] = ccddata.header['TAPLINES']
        # Add GAINMUL
        ccddata.header['GAINMUL'] = 1
        # Add CCDMODE
        if 'CDSSPEED' in ccddata.header:
            ccddata.header['CCDMODE'] = ccddata.header['CDSSPEED']
        else:
            ccddata.header['CCDMODE'] = 0
        # Add AMPMNUM
        if 'AMPMODE' in ccddata.header:
            ampmode = ccddata.header['AMPMODE']
            if ampmode == 'L2U2L1U1':
                ampnum = 0
            elif ampmode == 'L2U2':
                ampnum = 1
            elif ampmode == 'L1U1':
                ampnum = 2
            elif ampmode == 'L2L1':
                ampnum = 3
            elif ampmode == 'U2U1':
                ampnum = 4
            elif ampmode == 'L2':
                ampnum = 5
            elif ampmode == 'U2':
                ampnum = 6
            elif ampmode == 'L1':
                ampnum = 7
            elif ampmode == 'U1':
                ampnum = 8
            else:
                ampnum = 0
            ccddata.header['AMPMNUM'] = ampnum

            for amp in red_amp_dict.keys():
                if amp in ampmode:
                    # TODO: put in check for kw before updating
                    gkey = 'GAIN%d' % red_amp_dict[amp]
                    gain = red_amp_gain[amp]
                    ccddata.header[gkey] = gain
    else:
        print("ERROR -- illegal CAMERA keyword: %s" % ccddata.header['CAMERA'])
